import struct
import sys
import xml.etree.ElementTree as ET
from collections import deque
from collections.abc import Iterator
from dataclasses import dataclass
from typing import Any, Callable, Optional

import numpy as np

from ._exchange import Exchange, ExchangeGenerator, Exchanges
from .commands import Command, CommandException
from .responses import (
    Data,
    EndData,
    EndReason,
    FieldCapture,
    FrameData,
    ReadyData,
    StartData,
)

# Define the public API of this module
__all__ = [
    "NeedMoreData",
    "NoContextAvailable",
    "Buffer",
    "ControlConnection",
    "DataConnection",
]

# The names of the samples field used for averaging unscaled fields
# In newer versions it's GATE_DURATION but we keep SAMPLES for backwards
# compatibility
GATE_DURATION_FIELD = "PCAP.GATE_DURATION.Value"
SAMPLES_FIELD = "PCAP.SAMPLES.Value"


class NeedMoreData(Exception):
    """Raised if the `Buffer` isn't full enough to return the requested bytes"""


class NoContextAvailable(Exception):
    """Raised if there were no contexts available for this connection.
    This may result from calling `ControlConnection.receive_bytes()` without calling
    `ControlConnection.send()`, or if there were unmatched sends/receives"""


class Buffer:
    """Byte storage that provides line reader and bytes reader interfaces.
    For example::

        buf = Buffer()
        buf += bytes_from_server
        line = buf.read_line()  # raises NeedMoreData if no line
        for line in buf:
            pass
        bytes = buf.read_bytes(50)  # raises NeedMoreData if not enough bytes
    """

    def __init__(self):
        self._buf = bytearray()

    def __iadd__(self, byteslike: bytes):
        """Add some data from the server"""
        self._buf += byteslike
        return self

    def _extract_frame(self, num_to_extract, num_to_discard=0) -> bytearray:
        # extract num_to_extract bytes from the start of the buffer
        frame = self._buf[:num_to_extract]
        # Update the buffer in place, to take advantage of bytearray's
        # optimized delete-from-beginning feature.
        del self._buf[: num_to_extract + num_to_discard]
        return frame

    def read_bytes(self, num: int) -> bytearray:
        """Read and pop num bytes from the beginning of the buffer, raising
        `NeedMoreData` if the buffer isn't full enough to do so"""
        if num > len(self._buf):
            raise NeedMoreData()
        else:
            return self._extract_frame(num)

    def peek_bytes(self, num: int) -> bytearray:
        """Read but do not pop num bytes from the beginning of the buffer,
        raising `NeedMoreData` if the buffer isn't full enough to do so"""
        if num > len(self._buf):
            raise NeedMoreData()
        else:
            return self._buf[:num]

    def read_line(self):
        """Read and pop a newline terminated line (without terminator)
        from the beginning of the buffer, raising `NeedMoreData` if the
        buffer isn't full enough to do so"""
        idx = self._buf.find(b"\n")
        if idx < 0:
            raise NeedMoreData()
        else:
            return self._extract_frame(idx, num_to_discard=1)

    def __iter__(self):
        return self

    def __next__(self) -> bytes:
        try:
            return self.read_line()
        except NeedMoreData as err:
            raise StopIteration() from err


@dataclass
class _ExchangeContext:
    #: The exchange we should be filling
    exchange: Exchange
    #: The command that produced it
    command: Command
    #: If this was the last in the list, the generator to call next
    generator: Optional[ExchangeGenerator[Any]] = None

    def exception(self, e: Exception) -> CommandException:
        msg = f"{self.command} raised error:\n{type(e).__name__}: {e}"
        return CommandException(msg).with_traceback(e.__traceback__)


class ControlConnection:
    """Sans-IO connection to control port of PandA TCP server, supporting a
    Command based interface. For example::

        cc = ControlConnection()
        # Connection says what bytes should be sent to execute command
        to_send = cc.send(command)
        socket.sendall(to_send)
        while True:
            # Repeatedly process bytes from the PandA
            received = socket.recv()
            # Sending any subsequent bytes to be sent back to the PandA
            to_send = cc.receive_bytes(received)
            socket.sendall(to_send)
            # And processing the produced responses
            for command, response in cc.responses()
                do_something_with(response)
    """

    def __init__(self) -> None:
        self._buf = Buffer()
        self._lines: list[str] = []
        self._contexts: deque[_ExchangeContext] = deque()
        self._responses: deque[tuple[Command, Any]] = deque()

    def _update_contexts(self, lines: list[str], is_multiline=False) -> bytes:
        to_send = b""
        if len(self._contexts) == 0:
            raise NoContextAvailable()
        context = self._contexts.popleft()
        # Update the exchange with what we've got
        context.exchange.received = lines
        context.exchange.is_multiline = is_multiline
        # If we're given a generator to run then do so
        if context.generator:
            try:
                # Return the bytes from sending the next bit of the command
                exchanges = next(context.generator)
            except StopIteration as e:
                # Command complete, store the result
                self._responses.append((context.command, e.value))
            except Exception as e:
                # Command failed, store an exception
                self._responses.append((context.command, context.exception(e)))
            else:
                to_send = b"".join(
                    self._bytes_from_exchanges(
                        exchanges, context.command, context.generator
                    )
                )
        return to_send

    def _bytes_from_exchanges(
        self, exchanges: Exchanges, command: Command, generator: ExchangeGenerator[Any]
    ) -> Iterator[bytes]:
        if not isinstance(exchanges, list):
            exchanges = [exchanges]
        # No Exchanges when a Command's yield is empty e.g. unexpected/unparseable data
        # received from PandA
        if len(exchanges) == 0:
            return
        for ex in exchanges:
            context = _ExchangeContext(ex, command)
            self._contexts.append(context)
            text = "\n".join(ex.to_send) + "\n"
            yield text.encode()
        # The last exchange gets the generator so it triggers the next thing to send
        context.generator = generator

    def receive_bytes(self, received: bytes) -> bytes:
        """Tell the connection that you have received some bytes off the network.
        Parse these into high level responses which are yielded back by `responses`.
        Return any bytes to send back"""
        self._buf += received
        is_multiline = bool(self._lines)
        to_send = b""
        for line_b in self._buf:
            line = line_b.decode()
            if not is_multiline:
                # Check if we need to switch to multiline mode
                is_multiline = line.startswith("!") or line == "."
            if is_multiline:
                # Add a new line to the buffer
                self._lines.append(line)
                if line == ".":
                    # End of multiline mode, return what we've got
                    to_send += self._update_contexts(self._lines, is_multiline)
                    self._lines = []
                    is_multiline = False
                else:
                    # Check a correctly formatted response
                    assert line.startswith(
                        "!"
                    ), f"Multiline response {line} doesn't start with !"
            else:
                # Single line mode
                assert (
                    not self._lines
                ), f"Multiline response {self._lines} not terminated"
                to_send += self._update_contexts([line])
        return to_send

    def responses(self) -> Iterator[tuple[Command, Any]]:
        """Get the (command, response) tuples generated as part of the last
        receive_bytes"""
        while self._responses:
            yield self._responses.popleft()

    def send(self, command: Command) -> bytes:
        """Tell the connection you want to send an event, and it will return
        some bytes to send down the network
        """
        # If not given a partially run generator, start one here
        generator = command.execute()
        exchanges = next(generator)
        to_send = b"".join(self._bytes_from_exchanges(exchanges, command, generator))
        return to_send


class DataConnection:
    """Sans-IO connection to data port of PandA TCP server, supporting an
    flushable iterator interface. For example::

        dc = DataConnection()
        # Single connection string to send
        to_send = dc.connect()
        socket.sendall(to_send)
        while True:
            # Repeatedly process bytes from the PandA looking for data
            received = socket.recv()
            for data in dc.receive_bytes(received):
                do_something_with(data)
    """

    def __init__(self) -> None:
        # TODO: could support big endian, but are there any systems out there?
        assert sys.byteorder == "little", "PandA sends data little endian"
        # Store of bytes received so far to parse in the handlers
        self._buf = Buffer()
        # Header text from PandA with field info
        self._header = ""
        # The next parsing handler that should be called if there is data in buffer
        self._next_handler: Optional[Callable[[], Optional[Iterator[Data]]]] = None
        # numpy dtype of produced FrameData
        self._frame_dtype = None
        # frame data that has been received but not flushed yet
        self._partial_data = bytearray()
        # whether to flush after every frame
        self._flush_every_frame = False

    def _handle_connected(self):
        # Get the response from connect()
        line = self._buf.read_line()
        assert line == b"OK", f"Expected OK, got {line!r}"
        yield ReadyData()
        self._next_handler = self._handle_header_start

    def _handle_header_start(self):
        # Discard lines until we see header start tag
        line = self._buf.read_line()
        if line == b"<header>":
            self._header = line
            self._next_handler = self._handle_header_body

    def _handle_header_body(self):
        # Accumumlate header until the end tag, then parese and return
        line = self._buf.read_line()
        self._header += line
        if line == b"</header>":
            fields = []
            root = ET.fromstring(self._header)

            for field in root.find("fields"):
                fields.append(
                    FieldCapture(
                        name=str(field.get("name")),
                        type=np.dtype(field.get("type")),
                        capture=str(field.get("capture")),
                        scale=float(scale)
                        if (scale := field.get("scale")) is not None
                        else None,
                        offset=float(offset)
                        if (offset := field.get("offset")) is not None
                        else None,
                        units=str(units)
                        if (units := field.get("units")) is not None
                        else None,
                    )
                )
            data = root.find("data")
            sample_bytes = int(data.get("sample_bytes"))
            if sample_bytes - sum(f.type.itemsize for f in fields) == 4:
                # In raw mode with panda-server < 2.1 samples wasn't
                # put in if not specifically requested, but was still
                # sent
                name, capture = SAMPLES_FIELD.rsplit(".", maxsplit=1)
                fields.insert(
                    0,
                    FieldCapture(
                        name=name,
                        type=np.dtype("uint32"),
                        capture=capture,
                        scale=None,
                        offset=None,
                        units=None,
                    ),
                )
            self._frame_dtype = np.dtype(
                [(f"{f.name}.{f.capture}", f.type) for f in fields]
            )

            try:
                hw_time_offset_ns = np.int64(data.get("hw_time_offset_ns", ""))
            except ValueError:
                hw_time_offset_ns = None

            yield StartData(
                fields=fields,
                missed=int(data.get("missed")),
                process=str(data.get("process")),
                format=str(data.get("format")),
                sample_bytes=sample_bytes,
                arm_time=data.get("arm_time", None),
                start_time=data.get("start_time", None),
                hw_time_offset_ns=hw_time_offset_ns,
            )
            self._next_handler = self._handle_header_end

    def _handle_header_end(self):
        # Discard the newline at the end of the header
        assert self._buf.read_bytes(1) == b"\n", "Expected newline at end of header"
        self._next_handler = self._handle_data_start

    def _handle_data_start(self):
        # Handle "BIN " or "END "
        bytes = self._buf.read_bytes(4)
        if bytes == b"BIN ":
            self._next_handler = self._handle_data_frame
        elif bytes == b"END ":
            self._next_handler = self._handle_data_end
        else:
            raise ValueError(f"Bad data '{bytes}'")

    def _handle_data_frame(self):
        # Handle a whole data frame
        # Peek message length as uint32 LE
        # length = len("BIN " + 4_bytes_encoded_length + data)
        length = struct.unpack("<I", self._buf.peek_bytes(4))[0]
        # we already read "BIN ", so read the rest
        data = self._buf.read_bytes(length - 4)[4:]
        self._partial_data += data
        # if told to flush now, then yield what we have
        if self._flush_every_frame:
            yield from self.flush()
        self._next_handler = self._handle_data_start

    def _handle_data_end(self):
        # Handle the end reason
        samples, reason = self._buf.read_line().split(maxsplit=1)
        reason_enum = EndReason(reason.decode())
        # Flush whatever is not already flushed
        yield from self.flush()
        yield EndData(samples=int(samples), reason=reason_enum)
        self._next_handler = self._handle_header_start

    def receive_bytes(self, received: bytes, flush_every_frame=True) -> Iterator[Data]:
        """Tell the connection that you have received some bytes off the network.
        Parse these into Data structures and yield them back.

        Args:
            received: the bytes you received from the socket
            flush_every_frame: Whether to flush `FrameData` as soon as received.
                If False then they will only be sent if `flush` is called or
                end of acquisition reached
        """
        assert self._next_handler, "Connect not called"
        self._flush_every_frame = flush_every_frame
        self._buf += received
        while True:
            # Each of these handlers should call read at most once, so that
            # if we don't have enough data we don't lose partial data
            try:
                ret = self._next_handler()
                if ret:
                    # This is an iterator of Data objects
                    yield from ret
            except NeedMoreData:
                break

    def flush(self) -> Iterator[FrameData]:
        """If there is a partial data frame, pop and yield it"""
        if self._partial_data:
            # Make a numpy array wrapper to the bytearray, no copying here
            data = np.frombuffer(self._partial_data, self._frame_dtype)
            # Make a new bytearray, numpy view will keep the reference
            # to the old one so can't clear it in place
            self._partial_data = bytearray()
            yield FrameData(data)

    def connect(self, scaled: bool) -> bytes:
        """Return the bytes that need to be sent on connection"""
        assert not self._next_handler, "Can't connect while midway through collection"
        self._next_handler = self._handle_connected
        if scaled:
            return b"XML FRAMED SCALED\n"
        else:
            return b"XML FRAMED RAW\n"



import asyncio
import sys
import time

from pandablocks.asyncio import AsyncioClient
from pandablocks.commands import Arm, Put
from pandablocks.hdf import FrameProcessor, HDFWriter, create_pipeline, stop_pipeline
from pandablocks.responses import EndData, EndReason, FrameData, ReadyData


def print_progress_bar(fraction: float):
    # Print a simple progress bar, with a carriage return rather than newline
    # so that subsequent progress bars can be drawn on top
    print(f"{fraction * 100:5.1f}% [{'=' * int(fraction * 40):40s}]", end="\r")


async def hdf_queue_reporting():
    # Create the pipeline to scale and write HDF files
    pipeline = create_pipeline(
        FrameProcessor(), HDFWriter(scheme="/tmp/panda-capture-%d.h5")
    )
    try:
        async with AsyncioClient(sys.argv[1]) as client:
            # Gather data at 45MByte/s, should take about 60s
            repeats = 40000000
            await asyncio.gather(
                client.send(Put("SEQ1.REPEATS", repeats)),
                client.send(Put("SEQ1.PRESCALE", 0.5)),
            )
            progress = 0
            async for data in client.data(scaled=False, flush_period=1):
                # Always pass the data down the pipeline
                pipeline[0].queue.put_nowait(data)
                if isinstance(data, ReadyData):
                    # Data connection is ready, arm PandA
                    await client.send(Arm())
                elif isinstance(data, FrameData):
                    # Got some frame data, print a progress bar
                    progress += len(data.data)
                    print_progress_bar(progress / repeats)
                elif isinstance(data, EndData):
                    # We've done a single acquisition, check ok and return
                    assert data.reason == EndReason.OK, data.reason
                    break

    finally:
        start = time.time()
        print("\nClosing file...", end=" ")
        # Stop and wait for the pipeline to complete
        stop_pipeline(pipeline)
        print(f"took {time.time() - start:.1f} seconds")


if __name__ == "__main__":
    # One-shot run of a co-routine
    asyncio.run(hdf_queue_reporting())



import asyncio
import logging
from asyncio.streams import StreamReader, StreamWriter
from collections.abc import AsyncGenerator, Iterable
from contextlib import suppress
from typing import Optional

from .commands import Command, T
from .connections import ControlConnection, DataConnection
from .responses import Data

# Define the public API of this module
__all__ = ["AsyncioClient"]


class _StreamHelper:
    _reader: Optional[StreamReader] = None
    _writer: Optional[StreamWriter] = None

    @property
    def reader(self) -> StreamReader:
        assert self._reader, "connect() not called yet"
        return self._reader

    @property
    def writer(self) -> StreamWriter:
        assert self._writer, "connect() not called yet"
        return self._writer

    async def write_and_drain(self, data: bytes, timeout: Optional[float] = None):
        writer = self.writer
        writer.write(data)

        # Cannot simply await the drain, as if the remote end has disconnected
        # then the drain will never complete as the OS cannot clear its send buffer.
        write_task = asyncio.create_task(writer.drain())
        _, pending = await asyncio.wait([write_task], timeout=timeout)
        if len(pending):
            for task in pending:
                task.cancel()
            raise asyncio.TimeoutError("Timeout writing data")

    async def connect(self, host: str, port: int):
        self._reader, self._writer = await asyncio.open_connection(host, port)

    async def close(self):
        writer = self.writer
        self._reader = None
        self._writer = None
        writer.close()
        await writer.wait_closed()


class AsyncioClient:
    """Asyncio implementation of a PandABlocks client.
    For example::

        async with AsyncioClient("hostname-or-ip") as client:
            # Control port is now connected
            resp1, resp2 = await asyncio.gather(client.send(cmd1), client.send(cmd2))
            resp3 = await client.send(cmd3)
            async for data in client.data():
                handle(data)
        # Control and data ports are now disconnected
    """

    def __init__(self, host: str):
        self._host = host
        self._ctrl_connection = ControlConnection()
        self._ctrl_task: Optional[asyncio.Task] = None
        self._ctrl_queues: dict[int, asyncio.Queue] = {}
        self._ctrl_stream = _StreamHelper()

    async def connect(self):
        """Connect to the control port, and be ready to handle commands"""
        await self._ctrl_stream.connect(self._host, 8888)

        self._ctrl_task = asyncio.create_task(
            self._ctrl_read_forever(self._ctrl_stream.reader)
        )

    def is_connected(self):
        """True if there is a currently active connection.
        NOTE: This does not indicate if the remote end is still connected."""
        if self._ctrl_task and not self._ctrl_task.done():
            return True
        return False

    async def close(self):
        """Close the control connection, and wait for completion"""
        assert self._ctrl_task, "connect() not called yet"
        self._ctrl_task.cancel()
        await self._ctrl_stream.close()

    async def __aenter__(self) -> "AsyncioClient":
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def _ctrl_read_forever(self, reader: asyncio.StreamReader):
        """Continually read data from the stream reader and add to the data queue.

        Args:
            reader: The `StreamReader` to read from
        """
        while True:
            received = await reader.read(4096)
            if received == b"":
                error_str = (
                    "Received an empty packet. Closing connection. "
                    "Has the PandA disconnected?"
                )
                logging.error(error_str)
                raise ConnectionError(error_str)

            try:
                to_send = self._ctrl_connection.receive_bytes(received)
                await self._ctrl_stream.write_and_drain(to_send)
                for command, response in self._ctrl_connection.responses():
                    queue = self._ctrl_queues.pop(id(command))
                    queue.put_nowait(response)
            except Exception:
                logging.exception(f"Error handling '{received.decode()}'")

    async def send(self, command: Command[T], timeout: Optional[float] = None) -> T:
        """Send a command to control port of the PandA, returning its response.

        Args:
            command: The `Command` to send
        """
        queue: asyncio.Queue[T] = asyncio.Queue()
        # Need to use the id as non-frozen dataclasses don't hash
        self._ctrl_queues[id(command)] = queue
        to_send = self._ctrl_connection.send(command)
        await self._ctrl_stream.write_and_drain(to_send, timeout)
        response = await asyncio.wait_for(queue.get(), timeout)
        if isinstance(response, Exception):
            raise response
        else:
            return response

    async def data(
        self,
        scaled: bool = True,
        flush_period: Optional[float] = None,
        frame_timeout: Optional[float] = None,
    ) -> AsyncGenerator[Data, None]:
        """Connect to data port and yield data frames

        Args:
            scaled: Whether to scale and average data frames, reduces throughput
            flush_period: How often to flush partial data frames, None is on every
                chunk of data from the server
            frame_timeout: If no data is received for this amount of time, raise
                `asyncio.TimeoutError`
        """

        stream = _StreamHelper()
        connection = DataConnection()
        queue: asyncio.Queue[Iterable[Data]] = asyncio.Queue()

        def raise_timeouterror():
            raise asyncio.TimeoutError(f"No data received for {frame_timeout}s")
            yield

        async def periodic_flush():
            if flush_period is not None:
                while True:
                    # Every flush_period seconds flush and queue data
                    await asyncio.sleep(flush_period)
                    queue.put_nowait(connection.flush())

        async def read_from_stream():
            reader = stream.reader
            # Should we flush every FrameData?
            flush_every_frame = flush_period is None
            while True:
                try:
                    recv = await asyncio.wait_for(reader.read(4096), frame_timeout)
                except asyncio.TimeoutError:
                    queue.put_nowait(raise_timeouterror())
                    break
                else:
                    queue.put_nowait(connection.receive_bytes(recv, flush_every_frame))

        await stream.connect(self._host, 8889)
        await stream.write_and_drain(connection.connect(scaled))
        fut = asyncio.gather(periodic_flush(), read_from_stream())
        try:
            while True:
                for data in await queue.get():
                    yield data
        finally:
            fut.cancel()
            await stream.close()
            with suppress(asyncio.CancelledError):
                await fut



import asyncio
import os
import threading
from collections import deque
from collections.abc import Iterable, Iterator
from io import BufferedReader
from pathlib import Path

import numpy as np
import pytest
import pytest_asyncio

from pandablocks.connections import Buffer
from pandablocks.responses import (
    EndData,
    EndReason,
    FieldCapture,
    FrameData,
    ReadyData,
    StartData,
)


def chunked_read(f: BufferedReader, size: int) -> Iterator[bytes]:
    data = f.read(size)
    while data:
        yield data
        data = f.read(size)


@pytest_asyncio.fixture
def slow_dump():
    with open(Path(__file__).parent / "data_dumps/slow_dump.bin", "rb") as f:
        # Simulate small chunked read, sized so we hit the middle of a "BIN " marker
        yield chunked_read(f, 44)


@pytest_asyncio.fixture
def fast_dump():
    with open(Path(__file__).parent / "data_dumps/fast_dump.bin", "rb") as f:
        # Simulate larger chunked read
        yield chunked_read(f, 500)


@pytest_asyncio.fixture
def fast_dump_with_extra_header_params():
    """
    Add header parameters to `fast_dump.bin` binary stream. The fixture
    is used for testing absolute timing parameters passed in the header.
    """

    def fast_dump(extra_header_params):
        param_to_replace = 'sample_bytes="52"'
        params = " ".join(f'{k}="{v}"' for k, v in extra_header_params.items())
        params = " ".join([param_to_replace, params]) if params else param_to_replace
        with open(Path(__file__).parent / "data_dumps/fast_dump.bin", "rb") as f:
            # Simulate larger chunked read
            data = chunked_read(f, 500)
            for buffer in data:
                yield buffer.replace(param_to_replace.encode(), params.encode())

    return fast_dump


@pytest_asyncio.fixture
def raw_dump():
    with open(Path(__file__).parent / "data_dumps/raw_dump.bin", "rb") as f:
        # Simulate largest chunked read
        yield chunked_read(f, 200000)


@pytest_asyncio.fixture
def raw_dump_no_duration():
    with open(Path(__file__).parent / "data_dumps/raw_dump_no_duration.bin", "rb") as f:
        # Simulate largest chunked read
        yield chunked_read(f, 200000)


@pytest.fixture
def overrun_dump():
    with open(Path(__file__).parent / "data_dumps/raw_dump.bin", "rb") as f:
        # All in one go
        return f.read().replace(b"Disarmed", b"Data overrun")


DUMP_FIELDS = [
    FieldCapture(
        name="PCAP.BITS2",
        type=np.dtype("uint32"),
        capture="Value",
        scale=None,
        offset=None,
        units=None,
    ),
    FieldCapture(
        name="COUNTER1.OUT",
        type=np.dtype("float64"),
        capture="Min",
        scale=1,
        offset=0,
        units="",
    ),
    FieldCapture(
        name="COUNTER1.OUT",
        type=np.dtype("float64"),
        capture="Max",
        scale=1,
        offset=0,
        units="",
    ),
    FieldCapture(
        name="COUNTER3.OUT",
        type=np.dtype("float64"),
        capture="Value",
        scale=1,
        offset=0,
        units="",
    ),
    FieldCapture(
        name="PCAP.TS_START",
        type=np.dtype("float64"),
        capture="Value",
        scale=8e-09,
        offset=0,
        units="s",
    ),
    FieldCapture(
        name="COUNTER1.OUT",
        type=np.dtype("float64"),
        capture="Mean",
        scale=1,
        offset=0,
        units="",
    ),
    FieldCapture(
        name="COUNTER2.OUT",
        type=np.dtype("float64"),
        capture="Mean",
        scale=1,
        offset=0,
        units="",
    ),
]


def assert_all_data_in_hdf_file(hdf_file, samples_name):
    def multiples(num, offset=0):
        return pytest.approx(np.arange(1, 10001) * num + offset)

    assert hdf_file["/COUNTER1.OUT.Max"][:] == multiples(1)
    assert hdf_file["/COUNTER1.OUT.Mean"][:] == multiples(1)
    assert hdf_file["/COUNTER1.OUT.Min"][:] == multiples(1)
    assert hdf_file["/COUNTER2.OUT.Mean"][:] == multiples(2)
    assert hdf_file["/COUNTER3.OUT.Value"][:] == multiples(3)
    assert hdf_file["/PCAP.BITS2.Value"][:] == multiples(0)
    assert hdf_file[f"/{samples_name}"][:] == multiples(0, offset=125)
    assert hdf_file["/PCAP.TS_START.Value"][:] == multiples(2e-6, offset=7.2e-8 - 2e-6)


class Rows:
    def __init__(self, *rows):
        self.rows = rows

    def __eq__(self, o):
        same = o.tolist() == [pytest.approx(row) for row in self.rows]
        return same


@pytest_asyncio.fixture
def slow_dump_expected():
    yield [
        ReadyData(),
        StartData(DUMP_FIELDS, 0, "Scaled", "Framed", 52, None, None, None),
        FrameData(Rows([0, 1, 1, 3, 5.6e-08, 1, 2])),
        FrameData(Rows([8, 2, 2, 6, 1.000000056, 2, 4])),
        FrameData(Rows([0, 3, 3, 9, 2.000000056, 3, 6])),
        FrameData(Rows([8, 4, 4, 12, 3.000000056, 4, 8])),
        FrameData(Rows([0, 5, 5, 15, 4.000000056, 5, 10])),
        EndData(5, EndReason.DISARMED),
    ]


@pytest_asyncio.fixture
def fast_dump_expected():
    yield [
        ReadyData(),
        StartData(DUMP_FIELDS, 0, "Scaled", "Framed", 52, None, None, None),
        FrameData(
            Rows(
                [0, 1, 1, 3, 5.6e-08, 1, 2],
                [0, 2, 2, 6, 0.010000056, 2, 4],
                [8, 3, 3, 9, 0.020000056, 3, 6],
                [8, 4, 4, 12, 0.030000056, 4, 8],
                [8, 5, 5, 15, 0.040000056, 5, 10],
                [8, 6, 6, 18, 0.050000056, 6, 12],
                [8, 7, 7, 21, 0.060000056, 7, 14],
                [8, 8, 8, 24, 0.070000056, 8, 16],
                [8, 9, 9, 27, 0.080000056, 9, 18],
                [8, 10, 10, 30, 0.090000056, 10, 20],
            )
        ),
        FrameData(
            Rows(
                [0, 11, 11, 33, 0.100000056, 11, 22],
                [8, 12, 12, 36, 0.110000056, 12, 24],
                [8, 13, 13, 39, 0.120000056, 13, 26],
                [8, 14, 14, 42, 0.130000056, 14, 28],
                [8, 15, 15, 45, 0.140000056, 15, 30],
                [8, 16, 16, 48, 0.150000056, 16, 32],
                [8, 17, 17, 51, 0.160000056, 17, 34],
                [8, 18, 18, 54, 0.170000056, 18, 36],
                [8, 19, 19, 57, 0.180000056, 19, 38],
                [0, 20, 20, 60, 0.190000056, 20, 40],
                [8, 21, 21, 63, 0.200000056, 21, 42],
            )
        ),
        FrameData(
            Rows(
                [8, 22, 22, 66, 0.210000056, 22, 44],
                [8, 23, 23, 69, 0.220000056, 23, 46],
                [8, 24, 24, 72, 0.230000056, 24, 48],
                [8, 25, 25, 75, 0.240000056, 25, 50],
                [8, 26, 26, 78, 0.250000056, 26, 52],
                [8, 27, 27, 81, 0.260000056, 27, 54],
                [8, 28, 28, 84, 0.270000056, 28, 56],
                [0, 29, 29, 87, 0.280000056, 29, 58],
                [8, 30, 30, 90, 0.290000056, 30, 60],
                [8, 31, 31, 93, 0.300000056, 31, 62],
            )
        ),
        FrameData(
            Rows(
                [8, 32, 32, 96, 0.310000056, 32, 64],
                [8, 33, 33, 99, 0.320000056, 33, 66],
                [8, 34, 34, 102, 0.330000056, 34, 68],
                [8, 35, 35, 105, 0.340000056, 35, 70],
                [8, 36, 36, 108, 0.350000056, 36, 72],
                [8, 37, 37, 111, 0.360000056, 37, 74],
                [0, 38, 38, 114, 0.370000056, 38, 76],
                [8, 39, 39, 117, 0.380000056, 39, 78],
                [8, 40, 40, 120, 0.390000056, 40, 80],
                [8, 41, 41, 123, 0.400000056, 41, 82],
            )
        ),
        FrameData(
            Rows(
                [8, 42, 42, 126, 0.410000056, 42, 84],
                [8, 43, 43, 129, 0.420000056, 43, 86],
                [8, 44, 44, 132, 0.430000056, 44, 88],
                [8, 45, 45, 135, 0.440000056, 45, 90],
                [8, 46, 46, 138, 0.450000056, 46, 92],
                [0, 47, 47, 141, 0.460000056, 47, 94],
                [8, 48, 48, 144, 0.470000056, 48, 96],
                [8, 49, 49, 147, 0.480000056, 49, 98],
                [8, 50, 50, 150, 0.490000056, 50, 100],
                [8, 51, 51, 153, 0.500000056, 51, 102],
            )
        ),
        FrameData(
            Rows(
                [8, 52, 52, 156, 0.510000056, 52, 104],
                [8, 53, 53, 159, 0.520000056, 53, 106],
                [8, 54, 54, 162, 0.530000056, 54, 108],
                [8, 55, 55, 165, 0.540000056, 55, 110],
                [0, 56, 56, 168, 0.550000056, 56, 112],
                [8, 57, 57, 171, 0.560000056, 57, 114],
                [8, 58, 58, 174, 0.570000056, 58, 116],
            )
        ),
        EndData(58, EndReason.DISARMED),
    ]


class DummyServer:
    # Flag for useful debug output when writing tests
    # for diagnosing mismatching sent data.
    debug = False
    _debug_file = "out.txt"

    # Mechanism to tell the server to send a specific response back to the client
    # when it sees an expected string. When the expected message is seen the
    # response will be left-appended to the send buffer so it is sent next.
    # Items are removed from the Dict when they are sent.
    expected_message_responses: dict[str, str] = {}

    def __init__(self) -> None:
        # This will be added to whenever control port gets a message
        self.received: list[str] = []
        # Add to this to give the control port something to send back
        self.send: deque[str] = deque()
        # Add to this to give the data port something to send
        self.data: Iterable[bytes] = []

        if self.debug and os.path.isfile(self._debug_file):
            os.remove(self._debug_file)

    async def handle_ctrl(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        buf = Buffer()
        is_multiline = False

        while True:
            received = await reader.read(4096)
            if not received:
                break
            buf += received
            for line in buf:
                decoded_line = line.decode()
                self.received.append(decoded_line)
                if decoded_line in self.expected_message_responses:
                    self.send.appendleft(self.expected_message_responses[decoded_line])
                    del self.expected_message_responses[decoded_line]
                if line.endswith(b"<") or line.endswith(b"<B"):
                    is_multiline = True
                if not is_multiline or not line:
                    is_multiline = False
                    to_send = self.send.popleft() + "\n"
                    if self.debug:
                        with open(self._debug_file, "a") as f:
                            print(line, to_send, flush=True, file=f)
                    writer.write(to_send.encode())
                    await writer.drain()

    async def handle_data(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        # oneshot data writer
        await reader.read(4096)
        for data in self.data:
            await asyncio.sleep(0.1)
            writer.write(data)
            await writer.drain()

    async def open(self):
        self._ctrl_server = await asyncio.start_server(
            self.handle_ctrl, "127.0.0.1", 8888
        )
        self._data_server = await asyncio.start_server(
            self.handle_data, "127.0.0.1", 8889
        )

    async def close(self):
        self._ctrl_server.close()
        self._data_server.close()
        await self._ctrl_server.wait_closed()
        await self._data_server.wait_closed()


@pytest_asyncio.fixture
async def dummy_server_async():
    server = DummyServer()
    await server.open()
    yield server
    await server.close()


@pytest_asyncio.fixture
def dummy_server_in_thread():
    loop = asyncio.new_event_loop()
    server = DummyServer()
    t = threading.Thread(target=loop.run_forever)
    t.start()
    f = asyncio.run_coroutine_threadsafe(server.open(), loop)
    f.result()
    yield server
    asyncio.run_coroutine_threadsafe(server.close(), loop).result()
    loop.call_soon_threadsafe(loop.stop())
    t.join()


STATE_RESPONSES = [
    "!Attr=0\n.",
    "!Config=0\n.",
    "!Table<\n.",
    "!SingleLineMeta1=0\n!MultiLineMeta1<\n!SingleLineMeta2=0\n!MultiLineMeta2<\n.",
    "!tabledata\n.",
    "!multimetadata1\n.",
    "!multimetadata2\n.",
]

STATE_SAVEFILE = [
    "Attr=0",
    "Config=0",
    "SingleLineMeta1=0",
    "SingleLineMeta2=0",
    "Table<B",
    "tabledata",
    "",
    "MultiLineMeta1<",
    "multimetadata1",
    "",
    "MultiLineMeta2<",
    "multimetadata2",
    "",
]



import sys

import h5py
import pandas as pd

if __name__ == "__main__":
    with h5py.File(sys.argv[1], "r") as f:
        arm_time = f.attrs.get("arm_time", None)
        start_time = f.attrs.get("start_time", None)
        hw_time_offset_ns = f.attrs.get("hw_time_offset_ns", None)

        print(f"Arm time: {arm_time!r}")
        print(f"Start time: {start_time!r}")
        print(f"Hardware time offset: {hw_time_offset_ns!r} ns")

        if start_time:
            # Compute and print the start time that includes the offset
            ts_start = pd.Timestamp(start_time)
            if hw_time_offset_ns:
                ts_start += pd.Timedelta(nanoseconds=hw_time_offset_ns)
            print(f"Start time (system clock instead of hardware clock): {ts_start}")


# Expected output:
#
# Arm time: '2024-03-05T20:27:12.607841574Z'
# Start time: '2024-03-05T20:27:12.605729480Z'
# Hardware time offset: 2155797 ns
# Start TS including the offset: 2024-03-05T20:27:08.605729480Z



import asyncio
import copy

import pytest

from pandablocks.asyncio import AsyncioClient
from pandablocks.commands import CommandException, Get, Put

from .conftest import DummyServer

# Timeout in seconds
TIMEOUT = 3


@pytest.mark.asyncio
async def test_asyncio_get(dummy_server_async):
    dummy_server_async.send.append("OK =something")
    async with AsyncioClient("localhost") as client:
        response = await asyncio.wait_for(client.send(Get("PCAP.ACTIVE")), timeout=1)
    assert response == "something"
    assert dummy_server_async.received == ["PCAP.ACTIVE?"]


@pytest.mark.asyncio
async def test_asyncio_bad_put_raises(dummy_server_async):
    dummy_server_async.send.append("ERR no such field")
    async with AsyncioClient("localhost") as client:
        with pytest.raises(CommandException) as cm:
            await asyncio.wait_for(client.send(Put("PCAP.thing", 1)), timeout=1)
        assert (
            str(cm.value) == "Put(field='PCAP.thing', value=1) raised error:\n"
            "AssertionError: 'PCAP.thing=1' -> 'ERR no such field'"
        )
    assert dummy_server_async.received == ["PCAP.thing=1"]


@pytest.mark.asyncio
@pytest.mark.parametrize("disarmed", [True, False])
@pytest.mark.parametrize("flush_period", [0.1, None])
async def test_asyncio_data(
    dummy_server_async, fast_dump, fast_dump_expected, disarmed, flush_period
):
    if not disarmed:
        # simulate getting the data without the END marker as if arm was not pressed
        fast_dump = (x.split(b"END")[0] for x in fast_dump)
        fast_dump_expected = list(fast_dump_expected)[:-1]
    dummy_server_async.data = fast_dump
    events = []
    async with AsyncioClient("localhost") as client:
        async for data in client.data(frame_timeout=1, flush_period=flush_period):
            events.append(data)
            if len(events) == len(fast_dump_expected):
                break
    assert fast_dump_expected == events


@pytest.mark.parametrize(
    "timing_params",
    [
        {},
        {
            "arm_time": "2024-03-05T20:27:12.607841574Z",
            "start_time": "2024-03-05T20:27:12.608875498Z",
        },
        {
            "arm_time": "2024-03-05T20:27:12.607841574Z",
            "start_time": "2024-03-05T20:27:12.608875498Z",
            "hw_time_offset_ns": 100555,
        },
    ],
)
async def test_asyncio_data_with_abs_timing(
    dummy_server_async,
    fast_dump_with_extra_header_params,
    fast_dump_expected,
    timing_params,
):
    """
    The test for handling of `arm_time`, `start_time` and `hw_time_offset_ns`
    parameters passed in the header. The test is reusing the existing `fast_dump`
    and `fast_dump_expected` by adding timing parameters to the header in
    the binary stream and replacing the expected `StartData` attributes with
    the expected values.
    """
    dummy_server_async.data = fast_dump_with_extra_header_params(timing_params)
    events = []
    async with AsyncioClient("localhost") as client:
        async for data in client.data(frame_timeout=1):
            events.append(data)
            if len(events) == len(fast_dump_expected):
                break
    fast_dump_expected = list(fast_dump_expected)

    # Replace attributes in `StartData` with the expected values
    fast_dump_expected[1] = copy.deepcopy(fast_dump_expected[1])
    for attr_name in timing_params:
        setattr(fast_dump_expected[1], attr_name, timing_params[attr_name])

    assert fast_dump_expected == events


async def test_asyncio_data_timeout(dummy_server_async, fast_dump):
    dummy_server_async.data = fast_dump
    async with AsyncioClient("localhost") as client:
        with pytest.raises(asyncio.TimeoutError, match="No data received for 0.1s"):
            async for _ in client.data(frame_timeout=0.1):
                "This goes forever, when it runs out of data we will get our timeout"


async def test_asyncio_empty_frame_error():
    dummy_data = [b"ABC"] * 10 + [b""]
    dummy_data_iter = iter(dummy_data)

    async def dummy_read(n):
        return dummy_data_iter.__next__()

    reader = asyncio.StreamReader()
    reader.read = dummy_read

    written = []

    class DummyControlStream:
        async def write_and_drain(self, data):
            written.append(data)

    class DummyControlConnection:
        def receive_bytes(self, data):
            return data

    client = AsyncioClient("localhost")
    client._ctrl_stream = DummyControlStream()
    client._ctrl_connection = DummyControlConnection()
    with pytest.raises(
        ConnectionError, match="Received an empty packet. Closing connection."
    ):
        await client._ctrl_read_forever(reader)
    assert written == dummy_data[:-1]


@pytest.mark.asyncio
async def test_asyncio_connects(dummy_server_async: DummyServer):
    async with AsyncioClient("localhost") as client:
        assert client.is_connected()

    assert not client.is_connected()


@pytest.mark.asyncio
async def test_asyncio_client_uncontactable():
    """Test that a client raises an exception when the remote end is not
    contactable"""
    client = AsyncioClient("localhost")
    with pytest.raises(OSError):
        await client.connect()


@pytest.mark.asyncio
async def test_asyncio_client_fails_when_cannot_drain(dummy_server_async: DummyServer):
    """Test that we don't hang indefinitely when failing to drain data from the OS
    send buffer"""

    # Note this value is probably OS-dependant. I found it experimentally.
    large_data = b"ABC" * 100000000

    client = AsyncioClient("localhost")
    await client.connect()
    await dummy_server_async.close()
    with pytest.raises(asyncio.TimeoutError):
        await client._ctrl_stream.write_and_drain(large_data, timeout=TIMEOUT)

    # Can't use client.close() as it gets endlessly stuck. Do the important part.
    assert client._ctrl_task
    client._ctrl_task.cancel()



from collections import OrderedDict

import numpy as np
import pytest

from pandablocks.responses import TableFieldDetails, TableFieldInfo
from pandablocks.utils import UnpackedArray, table_to_words, words_to_table


@pytest.fixture
def table_fields() -> dict[str, TableFieldDetails]:
    """Table field definitions, taken from a SEQ.TABLE instance.
    Associated with table_data and table_field_info fixtures"""
    return {
        "REPEATS": TableFieldDetails(
            subtype="uint",
            bit_low=0,
            bit_high=15,
            description="Number of times the line will repeat",
            labels=None,
        ),
        "TRIGGER": TableFieldDetails(
            subtype="enum",
            bit_low=16,
            bit_high=19,
            description="The trigger condition to start the phases",
            labels=[
                "Immediate",
                "BITA=0",
                "BITA=1",
                "BITB=0",
                "BITB=1",
                "BITC=0",
                "BITC=1",
                "POSA>=POSITION",
                "POSA<=POSITION",
                "POSB>=POSITION",
                "POSB<=POSITION",
                "POSC>=POSITION",
                "POSC<=POSITION",
            ],
        ),
        "POSITION": TableFieldDetails(
            subtype="int",
            bit_low=32,
            bit_high=63,
            description="The position that can be used in trigger condition",
            labels=None,
        ),
        "TIME1": TableFieldDetails(
            subtype="uint",
            bit_low=64,
            bit_high=95,
            description="The time the optional phase 1 should take",
            labels=None,
        ),
        "OUTA1": TableFieldDetails(
            subtype="uint",
            bit_low=20,
            bit_high=20,
            description="Output A value during phase 1",
            labels=None,
        ),
        "OUTB1": TableFieldDetails(
            subtype="uint",
            bit_low=21,
            bit_high=21,
            description="Output B value during phase 1",
            labels=None,
        ),
        "OUTC1": TableFieldDetails(
            subtype="uint",
            bit_low=22,
            bit_high=22,
            description="Output C value during phase 1",
            labels=None,
        ),
        "OUTD1": TableFieldDetails(
            subtype="uint",
            bit_low=23,
            bit_high=23,
            description="Output D value during phase 1",
            labels=None,
        ),
        "OUTE1": TableFieldDetails(
            subtype="uint",
            bit_low=24,
            bit_high=24,
            description="Output E value during phase 1",
            labels=None,
        ),
        "OUTF1": TableFieldDetails(
            subtype="uint",
            bit_low=25,
            bit_high=25,
            description="Output F value during phase 1",
            labels=None,
        ),
        "TIME2": TableFieldDetails(
            subtype="uint",
            bit_low=96,
            bit_high=127,
            description="The time the mandatory phase 2 should take",
            labels=None,
        ),
        "OUTA2": TableFieldDetails(
            subtype="uint",
            bit_low=26,
            bit_high=26,
            description="Output A value during phase 2",
            labels=None,
        ),
        "OUTB2": TableFieldDetails(
            subtype="uint",
            bit_low=27,
            bit_high=27,
            description="Output B value during phase 2",
            labels=None,
        ),
        "OUTC2": TableFieldDetails(
            subtype="uint",
            bit_low=28,
            bit_high=28,
            description="Output C value during phase 2",
            labels=None,
        ),
        "OUTD2": TableFieldDetails(
            subtype="uint",
            bit_low=29,
            bit_high=29,
            description="Output D value during phase 2",
            labels=None,
        ),
        "OUTE2": TableFieldDetails(
            subtype="uint",
            bit_low=30,
            bit_high=30,
            description="Output E value during phase 2",
            labels=None,
        ),
        "OUTF2": TableFieldDetails(
            subtype="uint",
            bit_low=31,
            bit_high=31,
            description="Output F value during phase 2",
            labels=None,
        ),
    }


@pytest.fixture
def table_field_info(table_fields) -> TableFieldInfo:
    """Table data associated with table_fields and table_data fixtures"""
    return TableFieldInfo(
        "table", None, "Sequencer table of lines", 16384, table_fields, 4
    )


@pytest.fixture
def table_1_np_arrays() -> OrderedDict[str, UnpackedArray]:
    # Intentionally not in panda order. Whatever types the np arrays are,
    # the outputs from words_to_table will be uint32 or int32.
    return OrderedDict(
        {
            "REPEATS": np.array([5, 0, 50000], dtype=np.uint32),
            "TRIGGER": ["Immediate", "BITC=1", "Immediate"],
            "POSITION": np.array([-5, 678, 0], dtype=np.int32),
            "TIME1": np.array([100, 0, 9], dtype=np.uint32),
            "OUTA1": np.array([0, 1, 1], dtype=np.uint8),
            "OUTB1": np.array([0, 0, 1], dtype=np.uint8),
            "OUTD1": np.array([1, 0, 1], dtype=np.uint8),
            "OUTE1": np.array([0, 0, 1], dtype=np.uint8),
            "OUTC1": np.array([0, 0, 1], dtype=np.uint8),
            "OUTF1": np.array([1, 0, 1], dtype=np.uint8),
            "TIME2": np.array([0, 55, 9999], dtype=np.uint32),
            "OUTA2": np.array([0, 0, 1], dtype=np.uint8),
            "OUTB2": np.array([0, 0, 1], dtype=np.uint8),
            "OUTC2": np.array([1, 1, 1], dtype=np.uint8),
            "OUTD2": np.array([0, 0, 1], dtype=np.uint8),
            "OUTE2": np.array([0, 0, 1], dtype=np.uint8),
            "OUTF2": np.array([1, 0, 1], dtype=np.uint8),
        }
    )


@pytest.fixture
def table_1_np_arrays_int_enums() -> OrderedDict[str, UnpackedArray]:
    # Intentionally not in panda order. Whatever types the np arrays are,
    # the outputs from words_to_table will be uint32 or int32.
    return OrderedDict(
        {
            "POSITION": np.array([-5, 678, 0], dtype=np.int32),
            "TIME1": np.array([100, 0, 9], dtype=np.uint32),
            "OUTA1": np.array([0, 1, 1], dtype=np.uint8),
            "OUTB1": np.array([0, 0, 1], dtype=np.uint8),
            "OUTD1": np.array([1, 0, 1], dtype=np.uint8),
            "OUTE1": np.array([0, 0, 1], dtype=np.uint8),
            "OUTC1": np.array([0, 0, 1], dtype=np.uint8),
            "OUTF1": np.array([1, 0, 1], dtype=np.uint8),
            "TIME2": np.array([0, 55, 9999], dtype=np.uint32),
            "OUTA2": np.array([0, 0, 1], dtype=np.uint8),
            "OUTB2": np.array([0, 0, 1], dtype=np.uint8),
            "REPEATS": np.array([5, 0, 50000], dtype=np.uint32),
            "OUTC2": np.array([1, 1, 1], dtype=np.uint8),
            "OUTD2": np.array([0, 0, 1], dtype=np.uint8),
            "OUTE2": np.array([0, 0, 1], dtype=np.uint8),
            "OUTF2": np.array([1, 0, 1], dtype=np.uint8),
            "TRIGGER": np.array([0, 6, 0], dtype=np.uint8),
        }
    )


@pytest.fixture
def table_data_1() -> list[str]:
    return [
        "2457862149",
        "4294967291",
        "100",
        "0",
        "269877248",
        "678",
        "0",
        "55",
        "4293968720",
        "0",
        "9",
        "9999",
    ]


@pytest.fixture
def table_2_np_arrays() -> dict[str, UnpackedArray]:
    table: dict[str, UnpackedArray] = {
        "REPEATS": np.array([1, 0], dtype=np.uint32),
        "TRIGGER": ["Immediate", "Immediate"],
        "POSITION": np.array([-20, 2**31 - 1], dtype=np.int32),
        "TIME1": np.array([12, 2**32 - 1], dtype=np.uint32),
        "TIME2": np.array([32, 1], dtype=np.uint32),
    }

    table["OUTA1"] = np.array([0, 1], dtype=np.uint8)
    table["OUTA2"] = np.array([1, 0], dtype=np.uint8)
    for key in "BCDEF":
        table[f"OUT{key}1"] = table[f"OUT{key}2"] = np.array([0, 0], dtype=np.uint8)

    return table


@pytest.fixture
def table_data_2() -> list[str]:
    return [
        "67108865",
        "4294967276",
        "12",
        "32",
        "1048576",
        "2147483647",
        "4294967295",
        "1",
    ]


def test_table_packing_pack_length_mismatched(
    table_1_np_arrays: OrderedDict[str, UnpackedArray],
    table_field_info: TableFieldInfo,
):
    assert table_field_info.row_words

    # Adjust one of the record lengths so it mismatches
    field_info = table_field_info.fields[("OUTC1")]
    assert field_info
    table_1_np_arrays["OUTC1"] = np.array([1, 2, 3, 4, 5, 6, 7, 8])

    with pytest.raises(AssertionError):
        table_to_words(table_1_np_arrays, table_field_info)


@pytest.mark.parametrize(
    "table_fixture_name,table_data_fixture_name",
    [
        ("table_2_np_arrays", "table_data_2"),
        ("table_1_np_arrays", "table_data_1"),
    ],
)
def test_table_to_words_and_words_to_table(
    table_fixture_name: str,
    table_data_fixture_name: str,
    table_field_info: TableFieldInfo,
    request,
):
    table: dict[str, UnpackedArray] = request.getfixturevalue(table_fixture_name)
    table_data: list[str] = request.getfixturevalue(table_data_fixture_name)

    output_data = table_to_words(table, table_field_info)
    assert output_data == table_data
    output_table = words_to_table(
        output_data, table_field_info, convert_enum_indices=True
    )

    # Test the correct keys are outputted
    assert output_table.keys() == table.keys()

    # Check the items have been inserted in panda order
    sorted_output_table = OrderedDict({key: table[key] for key in output_table.keys()})
    assert sorted_output_table != OrderedDict(table)

    # Check the values are the same
    for output_key in output_table.keys():
        np.testing.assert_equal(output_table[output_key], table[output_key])


def test_table_packing_unpack(
    table_1_np_arrays: OrderedDict[str, np.ndarray],
    table_field_info: TableFieldInfo,
    table_data_1: list[str],
):
    assert table_field_info.row_words
    output_table = words_to_table(
        table_data_1, table_field_info, convert_enum_indices=True
    )

    actual: UnpackedArray
    for field_name, actual in output_table.items():
        expected = table_1_np_arrays[str(field_name)]
        np.testing.assert_array_equal(actual, expected)


def test_table_packing_unpack_no_convert_enum(
    table_1_np_arrays_int_enums: OrderedDict[str, UnpackedArray],
    table_field_info: TableFieldInfo,
    table_data_1: list[str],
):
    assert table_field_info.row_words
    output_table = words_to_table(table_data_1, table_field_info)

    actual: UnpackedArray
    for field_name, actual in output_table.items():
        expected = table_1_np_arrays_int_enums[str(field_name)]
        np.testing.assert_array_equal(actual, expected)


def test_table_packing_pack(
    table_1_np_arrays: dict[str, UnpackedArray],
    table_field_info: TableFieldInfo,
    table_data_1: list[str],
):
    assert table_field_info.row_words
    unpacked = table_to_words(table_1_np_arrays, table_field_info)

    for actual, expected in zip(unpacked, table_data_1):
        assert actual == expected


def test_table_packing_give_default_values(
    table_1_np_arrays: dict[str, UnpackedArray],
    table_field_info: TableFieldInfo,
):
    # We should have a complete table at the point of unpacking
    table_1_np_arrays["TRIGGER"] = np.array([])

    with pytest.raises(AssertionError) as e:
        table_to_words(table_1_np_arrays, table_field_info)
        assert (
            "Table record TRIGGER has mismatched length 0 compared "
            "to other records 3" in str(e.value)
        )



import asyncio
import io
import logging
import pathlib
from collections.abc import Awaitable

import click
from click.exceptions import ClickException

from pandablocks._control import interactive_control
from pandablocks.asyncio import AsyncioClient
from pandablocks.commands import GetState, SetState, T

# Default prompt
PROMPT = "< "
TUTORIAL = pathlib.Path(__file__).parent / "saves" / "tutorial.sav"


def asyncio_run(coro: Awaitable[T]) -> T:
    loop = asyncio.get_event_loop()
    try:
        return loop.run_until_complete(coro)
    finally:
        to_cancel = asyncio.tasks.all_tasks(loop)
        for task in to_cancel:
            task.cancel()
        loop.run_until_complete(asyncio.gather(*to_cancel, return_exceptions=True))


@click.group(invoke_without_command=True)
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(
        ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"], case_sensitive=False
    ),
)
@click.version_option()
@click.pass_context
def cli(ctx, log_level: str):
    """PandaBlocks client library command line interface."""

    level = getattr(logging, log_level.upper(), None)
    logging.basicConfig(format="%(levelname)s:%(message)s", level=level)

    # if no command is supplied, print the help message
    if ctx.invoked_subcommand is None:
        click.echo(cli.get_help(ctx))


@cli.command()
@click.option("--prompt", help="Prompt character", default=PROMPT, show_default=True)
@click.option(
    "--no-readline",
    help="Disable readline history and completion",
    is_flag=True,
)
@click.argument("host", type=str)
def control(host: str, prompt: str, no_readline: bool):
    """Open an interactive control console to HOST"""
    interactive_control(host, prompt, not no_readline)


@cli.command()
@click.argument("host")
@click.argument("outfile", type=click.File("w"))
def save(host: str, outfile: io.TextIOWrapper):
    """
    Save the current blocks configuration of HOST to OUTFILE
    """

    async def _save(host: str) -> list[str]:
        async with AsyncioClient(host) as client:
            return await client.send(GetState())

    state = asyncio_run(_save(host))
    outfile.write("\n".join(state) + "\n")


@cli.command()
@click.argument("host")
@click.argument("infile", type=click.File("r"), required=False)
@click.option("--tutorial", help="load the tutorial settings", is_flag=True)
def load(host: str, infile: io.TextIOWrapper, tutorial: bool):
    """
    Load a blocks configuration into HOST using the commands in INFILE
    """
    if tutorial:
        with TUTORIAL.open("r") as stream:
            state = stream.read().splitlines()
    elif infile is None:
        raise ClickException("INFILE not specified")
    else:
        state = infile.read().splitlines()

    async def _load(host: str, state: list[str]):
        async with AsyncioClient(host) as client:
            await client.send(SetState(state))

    asyncio_run(_load(host, state))


try:
    # Local import as we might not have h5py installed and want other commands
    # to work.
    from pandablocks.hdf import write_hdf_files

    @cli.command()
    @click.option(
        "--num",
        help="Number of collections to capture",
        default=1,
        show_default=True,
    )
    @click.option(
        "--arm",
        help="Arm PCAP at the start, and after each successful acquisition",
        is_flag=True,
    )
    @click.argument("host")
    @click.argument("scheme")
    def hdf(host: str, scheme: str, num: int, arm: bool):
        """
        Write an HDF file for each PCAP acquisition for HOST

        Uses the filename pattern specified by SCHEME, including %d for scan number
        starting from 1
        """

        async def _write_hdf_files(host: str, scheme: str, num: int, arm: bool):
            def file_name_generator(scheme: str):
                """Yield incrementally numbered file names based on provided scheme"""
                counter = 1
                while True:
                    yield scheme % counter
                    counter += 1

            async with AsyncioClient(host) as client:
                await write_hdf_files(
                    client, file_names=file_name_generator(scheme), num=num, arm=arm
                )

        # Don't use asyncio.run to workaround Python3.7 bug
        # https://bugs.python.org/issue38013
        asyncio_run(_write_hdf_files(host, scheme, num, arm))

except ImportError:

    @cli.command(hidden=True)
    def hdf():
        click.echo("ERROR: hdf subcommand unavailable - install 'hdf5' extras.")



import asyncio
import pprint
import sys

from pandablocks.asyncio import AsyncioClient
from pandablocks.commands import GetBlockInfo, GetFieldInfo, GetPcapBitsLabels


async def introspect():
    # Create a client and connect the control and data ports
    async with AsyncioClient(sys.argv[1]) as client:
        # Get the list of all blocks in the PandA
        block_info = await client.send(GetBlockInfo())
        # Find and print all fields for each block
        for block in block_info:
            field_info = await client.send(GetFieldInfo(block))
            pprint.pprint({block: field_info})

        # Get the labels for every PCAP.BITS[n] fields
        labels = await client.send(GetPcapBitsLabels())
        pprint.pprint(labels)


if __name__ == "__main__":
    # One-shot run of a co-routine
    asyncio.run(introspect())



from collections.abc import Iterable
from typing import Union, cast

import numpy as np
import numpy.typing as npt

from pandablocks.responses import TableFieldInfo

UnpackedArray = Union[
    npt.NDArray[np.uint8],
    npt.NDArray[np.uint16],
    npt.NDArray[np.int32],
    npt.NDArray[np.uint32],
    list[str],
]


def words_to_table(
    words: Iterable[str],
    table_field_info: TableFieldInfo,
    convert_enum_indices: bool = False,
) -> dict[str, UnpackedArray]:
    """Unpacks the given `packed` data based on the fields provided.
    Returns the unpacked data in {column_name: column_data} column-indexed format

    Args:
        words: An iterable of data for this table, from PandA. Each item is
            expected to be the string representation of a uint32.
        table_fields_info: The info for tables, containing the number of words per row,
            and the bit information for fields.
        convert_enum_indices: If True, convert all enum values to their string
            representation. Otherwise return enums as integer values
    Returns:
        unpacked: A dict containing record information, where keys are field names
            and values are numpy arrays or a sequence of strings of record values
            in that column.
    """

    row_words = table_field_info.row_words
    data = np.array(words, dtype=np.uint32)
    # Convert 1-D array into 2-D, one row element per row in the PandA table
    data = data.reshape(len(data) // row_words, row_words)
    packed = data.T

    unpacked: dict[str, UnpackedArray] = {}

    for field_name, field_info in table_field_info.fields.items():
        offset = field_info.bit_low
        bit_length = field_info.bit_high - field_info.bit_low + 1

        # The word offset indicates which column this field is in
        # (column is exactly one 32-bit word)
        word_offset = offset // 32

        # bit offset is location of field inside the word
        bit_offset = offset & 0x1F

        # Mask to remove every bit that isn't in the range we want
        mask = (1 << bit_length) - 1

        value: npt.NDArray[np.uint32] = (packed[word_offset] >> bit_offset) & mask
        packing_value: UnpackedArray

        if field_info.subtype == "int":
            # First convert from 2's complement to offset, then add in offset.
            temp = (value.astype(np.int64) ^ (1 << (bit_length - 1))) + (
                -1 << (bit_length - 1)
            )
            packing_value = temp.astype(np.int32)
        elif field_info.subtype == "enum" and convert_enum_indices:
            assert field_info.labels, f"Enum field {field_name} has no labels"
            packing_value = [field_info.labels[x] for x in value]
        else:
            # Use shorter types, as these are used in waveform creation
            if bit_length <= 8:
                packing_value = value.astype(np.uint8)
            elif bit_length <= 16:
                packing_value = value.astype(np.uint16)
            else:
                packing_value = value.astype(np.uint32)  # already uint32

        unpacked.update({field_name: packing_value})

    return unpacked


def table_to_words(
    table: dict[str, UnpackedArray], table_field_info: TableFieldInfo
) -> list[str]:
    """Convert records based on the field definitions into the format PandA expects
    for table writes.

    Args:
        table: A dict containing record information, where keys are field names
            and values are iterables of record values in that column.
        table_field_info: The info for tables, containing the dict `fields` for
            information on each field, and the number of words per row.
    Returns:
        list[str]: The list of data ready to be sent to PandA
    """
    row_words = table_field_info.row_words

    # Iterate over the zipped fields and their associated records to construct the
    # packed array.
    packed = None

    for column_name, column in table.items():
        field_details = table_field_info.fields[column_name]
        if field_details.labels and len(column) and isinstance(column[0], str):
            # Must convert the list of strings to list of ints
            column_value = np.array(
                [field_details.labels.index(x) for x in column], dtype=np.uint32
            )
        else:
            # PandA always handles tables in uint32 format
            column_value = np.array(column, dtype=np.uint32)

        if packed is None:
            # Create 1-D array sufficiently long to exactly hold the entire table, cast
            # to prevent type error, this will still work if column is another iterable
            # e.g numpy array
            packed = np.zeros((len(column), row_words), dtype=np.uint32)
        else:
            assert len(packed) == len(column), (
                f"Table record {column_name} has mismatched length {len(column)} "
                f"compared to other records {len(packed)}, cannot pack data. "
                "If setting values through the ioc, ensure all values are given "
                "before submitting."
            )

        offset = field_details.bit_low

        # The word offset indicates which column this field is in
        # (each column is one 32-bit word)
        word_offset = offset // 32
        # bit offset is location of field inside the word
        bit_offset = offset & 0x1F

        # Slice to get the column to apply the values to.
        # bit shift the value to the relevant bits of the word

        packed[:, word_offset] |= cast(np.unsignedinteger, column_value) << bit_offset

    assert isinstance(packed, np.ndarray), "Table has no columns"  # Squash mypy warning

    # 2-D array -> 1-D array -> list[int] -> list[str]
    return [str(x) for x in packed.flatten().tolist()]



import asyncio
import sys

from pandablocks.asyncio import AsyncioClient
from pandablocks.commands import Put
from pandablocks.hdf import write_hdf_files


async def arm_and_hdf():
    # Create a client and connect the control and data ports
    async with AsyncioClient(sys.argv[1]) as client:
        # Put to 2 fields simultaneously
        await asyncio.gather(
            client.send(Put("SEQ1.REPEATS", 1000)),
            client.send(Put("SEQ1.PRESCALE", 1000)),
        )
        # Listen for data, arming the PandA at the beginning
        await write_hdf_files(client, scheme="/tmp/panda-capture-%d.h5", arm=True)


if __name__ == "__main__":
    # One-shot run of a co-routine
    asyncio.run(arm_and_hdf())



from string import digits
from typing import Optional

from .blocking import BlockingClient
from .commands import FieldInfo, GetBlockInfo, GetFieldInfo, Raw, is_multiline_command

# Timeout for commands to PandA
TIMEOUT = 2


STATIC_STAR_COMMANDS = [
    "*IDN?",  # Device identification.
    "*WHO?",  # List connected clients.
    "*BLOCKS?",  # List device blocks.
    "*CAPTURE?",  # Report data capture words.
    "*POSITIONS?",  # Enumerate possible capture positions.
    "*VERBOSE=",  # Control command logging.
    "*SAVESTATE=",  # Triggers immediate save to file of the persistence file state.
]

for suff in ("STATUS?", "CAPTURED?", "COMPLETION?", "ARM=", "DISARM="):
    STATIC_STAR_COMMANDS.append(f"*PCAP.{suff}")  # Position capture status/action
for suff in ("", ".CONFIG", ".BITS", ".POSN", ".READ", ".ATTR", ".TABLE"):
    STATIC_STAR_COMMANDS.append(f"*CHANGES{suff}?")  # Report changes to values
    STATIC_STAR_COMMANDS.append(f"*CHANGES{suff}=")  # Reset reported changes


def _get_user_input(prompt) -> list[str]:
    lines = [input(prompt)]
    if is_multiline_command(lines[0]):
        while lines[-1]:
            lines.append(input(prompt))
    return lines


def text_matches(t1, t2):
    return t1.startswith(t2) or t2.startswith(t1)


class BlockCompleter:
    def __init__(self, client: BlockingClient):
        self.matches: list[str] = []
        self._client = client
        self._blocks = self._client.send(
            GetBlockInfo(skip_description=True), timeout=TIMEOUT
        )
        self._fields = self._get_fields(list(self._blocks))
        # TODO: Extend use of _fields now we have enum labels available?

    def _get_fields(self, blocks: list[str]) -> dict[str, dict[str, FieldInfo]]:
        fields = self._client.send(
            [GetFieldInfo(block, extended_metadata=False) for block in blocks],
            timeout=TIMEOUT,
        )
        return dict(zip(blocks, fields))

    def _with_suffixes(self, block: str, numbers: bool) -> list[str]:
        block_info = self._blocks[block]
        num = block_info.number
        if numbers and num > 1:
            return [f"{block}{i}" for i in range(1, num + 1)]
        else:
            return [block]

    def _block_field_matches(self, text: str, prefix="") -> list[str]:
        matches = []
        text = text[len(prefix) :]
        split = text.split(".", maxsplit=1)
        if len(split) == 1:
            # No dot, text is a partial block name
            for block in self._blocks:
                if block.startswith(text):
                    for bn in self._with_suffixes(block, numbers=not prefix):
                        if bn.startswith(text):
                            matches.append(prefix + bn)
        else:
            bn = split[0]
            for field in self._fields[bn.rstrip(digits)]:
                bnf = f"{bn}.{field}"
                if bnf.startswith(text):
                    matches.append(prefix + bnf)
        return matches

    def __call__(self, text: str, state: int) -> Optional[str]:
        """Return the next possible completion for 'text'.
        This is called successively with state == 0, 1, 2, ... until it
        returns None.  The completion should begin with 'text'.
        """
        if state == 0:
            # Create match cache
            if text.startswith("*"):
                # Star command completions
                self.matches = [
                    x for x in STATIC_STAR_COMMANDS if text_matches(text, x)
                ]
                if not self.matches:
                    # *DESC and *ENUMS have block names in them
                    if text_matches(text, "*DESC"):
                        self.matches = self._block_field_matches(text, "*DESC.")
                    elif text_matches(text, "*ENUMS"):
                        self.matches = self._block_field_matches(text, "*ENUMS.")
            else:
                # Check for block names
                self.matches = self._block_field_matches(text)
            if len(self.matches) == 1 and len(text) > len(self.matches[0]):
                # We only partially match, if the text is longer then don't complete
                self.matches = []
        # Return correct cached match
        try:
            return self.matches[state]
        except IndexError:
            return None


def set_completer(completer):
    try:
        # Give command completion etc.
        import readline
    except ImportError:
        # Don't need it on windows
        pass
    else:
        readline.set_completer(completer)
        # Apple does not ship GNU readline with OS X.
        # It does ship BSD libedit which includes a readline compatibility interface.
        # https://stackoverflow.com/a/7116997
        if "libedit" in readline.__doc__:
            readline.parse_and_bind("bind ^I rl_complete")
        else:
            readline.parse_and_bind("tab: complete")
        # Only complete at the start of the line, not for values
        readline.set_completer_delims("")


def interactive_control(host: str, prompt: str, readline=True):
    client = BlockingClient(host)
    client.connect()
    try:
        if readline:
            # Complete Block names when tab key pressed
            set_completer(BlockCompleter(client))
        while True:
            command = Raw(_get_user_input(prompt))
            for line in client.send(command):
                print(line)
    except (EOFError, KeyboardInterrupt):
        # End with a newline
        print()
    finally:
        client.close()



import logging
import re
from collections.abc import Generator
from dataclasses import dataclass, field
from enum import Enum
from typing import (
    Any,
    Callable,
    Generic,
    Optional,
    TypeVar,
    Union,
    overload,
)

from ._exchange import Exchange, ExchangeGenerator
from .responses import (
    BitMuxFieldInfo,
    BitOutFieldInfo,
    BlockInfo,
    Changes,
    EnumFieldInfo,
    ExtOutBitsFieldInfo,
    ExtOutFieldInfo,
    FieldInfo,
    PosMuxFieldInfo,
    PosOutFieldInfo,
    ScalarFieldInfo,
    SubtypeTimeFieldInfo,
    TableFieldDetails,
    TableFieldInfo,
    TimeFieldInfo,
    UintFieldInfo,
)

# Define the public API of this module
__all__ = [
    "Command",
    "CommandException",
    "Raw",
    "Get",
    "GetLine",
    "GetMultiline",
    "Put",
    "Append",
    "Arm",
    "Disarm",
    "GetBlockInfo",
    "GetFieldInfo",
    "GetPcapBitsLabels",
    "ChangeGroup",
    "GetChanges",
    "GetState",
    "SetState",
]


T = TypeVar("T")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")


# Checks whether the server will interpret cmd as a table command: search for
# first of '?', '=', '<', if '<' found first then it's a multiline command.
MULTILINE_COMMAND = re.compile(r"^[^?=]*<")


def is_multiline_command(cmd: str):
    return MULTILINE_COMMAND.match(cmd) is not None


@dataclass
class Command(Generic[T]):
    """Abstract baseclass for all ControlConnection commands to be inherited from"""

    def execute(self) -> ExchangeGenerator[T]:
        # A generator that sends lines to the PandA, gets lines back, and returns a
        # response
        raise NotImplementedError(self)


class CommandException(Exception):
    """Raised if a `Command` receives a mal-formed response"""


# `execute_commands()` actually returns a list with length equal to the number
# of tasks passed; however, Tuple is used similar to the annotation for
# zip() because typing does not support variadic type variables.  See
# typeshed PR #1550 for discussion.
@overload
def _execute_commands(c1: Command[T]) -> ExchangeGenerator[tuple[T]]: ...


@overload
def _execute_commands(
    c1: Command[T], c2: Command[T2]
) -> ExchangeGenerator[tuple[T, T2]]: ...


@overload
def _execute_commands(
    c1: Command[T], c2: Command[T2], c3: Command[T3]
) -> ExchangeGenerator[tuple[T, T2, T3]]: ...


@overload
def _execute_commands(
    c1: Command[T], c2: Command[T2], c3: Command[T3], c4: Command[T4]
) -> ExchangeGenerator[tuple[T, T2, T3, T4]]: ...


@overload
def _execute_commands(
    *commands: Command[Any],
) -> ExchangeGenerator[tuple[Any, ...]]: ...


def _execute_commands(*commands):
    """Call the `Command.execute` method on each of the commands to produce
    some `Exchange` generators, which are yielded back to the connection,
    then zip together the responses to those exchanges into a tuple"""
    # If we add type annotations to this function then mypy complains:
    # Overloaded function implementation does not accept all possible arguments
    # As we want to type check this, we put the logic in _zip_with_return
    ret = yield from _zip_with_return([command.execute() for command in commands])
    return ret


def _zip_with_return(
    generators: list[ExchangeGenerator[Any]],
) -> ExchangeGenerator[tuple[Any, ...]]:
    # Sentinel to show what generators are not yet exhausted
    pending = object()
    returns = [pending] * len(generators)
    while True:
        yields: list[Exchange] = []
        for i, gen in enumerate(generators):
            # If we haven't exhausted the generator
            if returns[i] is pending:
                try:
                    # Get the exchanges that it wants to fill in
                    exchanges = next(gen)
                except StopIteration as e:
                    # Generator is exhausted, store its return value
                    returns[i] = e.value
                else:
                    # Add the exchanges to the list
                    if isinstance(exchanges, list):
                        yields += exchanges
                    else:
                        yields.append(exchanges)
        if yields:
            # There were some Exchanges yielded, so yield them all up
            # for the Connection to fill in
            yield yields
        else:
            # All the generators are exhausted, so return the tuple of all
            # their return values
            return tuple(returns)


@dataclass
class Raw(Command[list[str]]):
    """Send a raw command

    Args:
        inp: The input lines to send

    For example::

        Raw(["PCAP.ACTIVE?"]) -> ["OK =1"]
        Raw(["SEQ1.TABLE>", "1", "1", "0", "0", ""]) -> ["OK"]
        Raw(["SEQ1.TABLE?"]) -> ["!1", "!1", "!0", "!0", "."])
    """

    inp: list[str]

    def execute(self) -> ExchangeGenerator[list[str]]:
        ex = Exchange(self.inp)
        yield ex
        return ex.received


@dataclass
class Get(Command[Union[str, list[str]]]):
    """Get the value of a field or star command.

    If the form of the expected return is known, consider using `GetLine`
    or `GetMultiline` instead.

    Args:
        field: The field, attribute, or star command to get

    For example::

        Get("PCAP.ACTIVE") -> "1"
        Get("SEQ1.TABLE") -> ["1048576", "0", "1000", "1000"]
        Get("*IDN") -> "PandA 1.1..."
    """

    field: str

    def execute(self) -> ExchangeGenerator[Union[str, list[str]]]:
        ex = Exchange(f"{self.field}?")
        yield ex
        if ex.is_multiline:
            return ex.multiline
        else:
            return ex.line


@dataclass
class GetLine(Command[str]):
    """Get the value of a field or star command, when the result is expected to be a
    single line.

    Args:
        field: The field, attribute, or star command to get

    For example::

        GetLine("PCAP.ACTIVE") -> "1"
        GetLine("*IDN") -> "PandA 1.1..."
    """

    field: str

    def execute(self) -> ExchangeGenerator[str]:
        ex = Exchange(f"{self.field}?")
        yield ex
        return ex.line


@dataclass
class GetMultiline(Command[list[str]]):
    """Get the value of a field or star command, when the result is expected to be a
    multiline response.

    Args:
        field: The field, attribute, or star command to get

    For example::

        GetMultiline("SEQ1.TABLE") -> ["1048576", "0", "1000", "1000"]
        GetMultiline("*METADATA.*") -> ["LABEL_FILTER1", "APPNAME", ...]
    """

    field: str

    def execute(self) -> ExchangeGenerator[list[str]]:
        ex = Exchange(f"{self.field}?")
        yield ex
        return ex.multiline


@dataclass
class Put(Command[None]):
    """Put the value of a field.

    Args:
        field: The field, attribute, or star command to put
        value: The value, either string or list of strings or empty, to put

    For example::

        Put("PCAP.TRIG", "PULSE1.OUT")
        Put("SEQ1.TABLE", ["1048576", "0", "1000", "1000"])
        Put("SFP3_SYNC_IN1.SYNC_RESET")
    """

    field: str
    value: Union[str, list[str]] = ""

    def execute(self) -> ExchangeGenerator[None]:
        if isinstance(self.value, list):
            # Multiline table with blank line to terminate
            ex = Exchange([f"{self.field}<"] + self.value + [""])
        else:
            ex = Exchange(f"{self.field}={self.value}")
        yield ex
        ex.check_ok()


@dataclass
class Append(Command[None]):
    """Append the value of a table field.

    Args:
        field: The field, attribute, or star command to append
        value: The value, list of strings, to append

    For example::

        Append("SEQ1.TABLE", ["1048576", "0", "1000", "1000"])
    """

    field: str
    value: list[str]

    def execute(self) -> ExchangeGenerator[None]:
        # Multiline table with blank line to terminate
        ex = Exchange([f"{self.field}<<"] + self.value + [""])
        yield ex
        ex.check_ok()


class Arm(Command[None]):
    """Arm PCAP for an acquisition by sending ``*PCAP.ARM=``"""

    def execute(self) -> ExchangeGenerator[None]:
        ex = Exchange("*PCAP.ARM=")
        yield ex
        ex.check_ok()


class Disarm(Command[None]):
    """Disarm PCAP, stopping acquisition by sending ``*PCAP.DISARM=``"""

    def execute(self) -> ExchangeGenerator[None]:
        ex = Exchange("*PCAP.DISARM=")
        yield ex
        ex.check_ok()


@dataclass
class GetBlockInfo(Command[dict[str, BlockInfo]]):
    """Get the name, number, and description of each block type
    in a dictionary, alphabetically ordered

    Args:
        skip_description: If `True`, prevents retrieving the description
            for each Block. This will reduce network calls.

    For example::

         GetBlockInfo() ->
             {
                 "LUT": BlockInfo(number=8, description="Lookup table"),
                 "PCAP": BlockInfo(number=1, description="Position capture control"),
                 ...
             }
    """

    skip_description: bool = False

    def execute(self) -> ExchangeGenerator[dict[str, BlockInfo]]:
        ex = Exchange("*BLOCKS?")
        yield ex

        blocks_list, commands = [], []
        for line in ex.multiline:
            block, num = line.split()
            blocks_list.append((block, int(num)))
            commands.append(GetLine(f"*DESC.{block}"))

        if self.skip_description:
            # Must use tuple() to match type returned by _execute_commands
            description_values = tuple(None for _ in commands)
        else:
            description_values = yield from _execute_commands(*commands)

        block_infos = {
            block: BlockInfo(number=num, description=desc)
            for (block, num), desc in sorted(zip(blocks_list, description_values))
        }

        return block_infos


# The type of the generators used for creating the Get commands for each field
# and setting the returned data into the FieldInfo structure
_FieldGeneratorType = Generator[
    Union[Exchange, list[Exchange]],
    None,
    tuple[str, FieldInfo],
]


@dataclass
class GetFieldInfo(Command[dict[str, FieldInfo]]):
    """Get the fields of a block, returning a `FieldInfo` (or appropriate subclass) for
    each one, ordered to match the definition order in the PandA

    Args:
        block: The name of the block type
        extended_metadata: If `True`, retrieves detailed metadata about a field and
            all of its attributes. This will cause an additional network round trip.
            If `False` only the field names and types will be returned. Default `True`.

    For example::

        GetFieldInfo("LUT") -> {
            "INPA":
                BitMuxFieldInfo(type='bit_mux',
                                subtype=None,
                                description='Input A',
                                max_delay=5
                                label=['TTLIN1.VAL', 'TTLIN2.VAL', ...]),
            ...}
    """

    block: str
    extended_metadata: bool = True

    _commands_map: dict[
        tuple[str, Optional[str]],
        Callable[
            [str, str, Optional[str]],
            _FieldGeneratorType,
        ],
    ] = field(init=False, repr=False, default_factory=dict)

    def __post_init__(self):
        # Map a (type, subtype) to a method that returns the appropriate
        # subclasss of FieldInfo, and a list of all the Commands to request.
        # Note that fields that do not have additional attributes are not listed.
        self._commands_map = {
            # Order matches that of PandA server's Field Types docs
            ("time", None): self._time,
            ("bit_out", None): self._bit_out,
            ("pos_out", None): self._pos_out,
            ("ext_out", "timestamp"): self._ext_out,
            ("ext_out", "samples"): self._ext_out,
            ("ext_out", "bits"): self._ext_out_bits,
            ("bit_mux", None): self._bit_mux,
            ("pos_mux", None): self._pos_mux,
            ("table", None): self._table,
            ("param", "uint"): self._uint,
            ("read", "uint"): self._uint,
            ("write", "uint"): self._uint,
            ("param", "int"): self._no_attributes,
            ("read", "int"): self._no_attributes,
            ("write", "int"): self._no_attributes,
            ("param", "scalar"): self._scalar,
            ("read", "scalar"): self._scalar,
            ("write", "scalar"): self._scalar,
            ("param", "bit"): self._no_attributes,
            ("read", "bit"): self._no_attributes,
            ("write", "bit"): self._no_attributes,
            ("param", "action"): self._no_attributes,
            ("read", "action"): self._no_attributes,
            ("write", "action"): self._no_attributes,
            ("param", "lut"): self._no_attributes,
            ("read", "lut"): self._no_attributes,
            ("write", "lut"): self._no_attributes,
            ("param", "enum"): self._enum,
            ("read", "enum"): self._enum,
            ("write", "enum"): self._enum,
            ("param", "time"): self._subtype_time,
            ("read", "time"): self._subtype_time,
            ("write", "time"): self._subtype_time,
        }

    def _get_desc(self, field_name: str) -> GetLine:
        """Create the Command to retrieve the description"""
        return GetLine(f"*DESC.{self.block}.{field_name}")

    def _uint(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        desc, maximum = yield from _execute_commands(
            self._get_desc(field_name),
            GetLine(f"{self.block}1.{field_name}.MAX"),
        )

        field_info = UintFieldInfo(field_type, field_subtype, desc, int(maximum))
        return field_name, field_info

    def _scalar(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        desc, units, scale, offset = yield from _execute_commands(
            self._get_desc(field_name),
            GetLine(f"{self.block}.{field_name}.UNITS"),
            GetLine(f"{self.block}.{field_name}.SCALE"),
            GetLine(f"{self.block}.{field_name}.OFFSET"),
        )

        field_info = ScalarFieldInfo(
            field_type, field_subtype, desc, units, float(scale), int(offset)
        )

        return field_name, field_info

    def _subtype_time(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        desc, units_labels = yield from _execute_commands(
            self._get_desc(field_name),
            GetMultiline(f"*ENUMS.{self.block}.{field_name}.UNITS"),
        )

        field_info = SubtypeTimeFieldInfo(field_type, field_subtype, desc, units_labels)

        return field_name, field_info

    def _enum(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        desc, labels = yield from _execute_commands(
            self._get_desc(field_name),
            GetMultiline(f"*ENUMS.{self.block}.{field_name}"),
        )

        field_info = EnumFieldInfo(field_type, field_subtype, desc, labels)

        return field_name, field_info

    def _time(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        desc, units = yield from _execute_commands(
            self._get_desc(field_name),
            GetMultiline(f"*ENUMS.{self.block}.{field_name}.UNITS"),
        )

        field_info = TimeFieldInfo(field_type, field_subtype, desc, units)

        return field_name, field_info

    def _bit_out(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        desc, capture_word, offset = yield from _execute_commands(
            self._get_desc(field_name),
            GetLine(f"{self.block}1.{field_name}.CAPTURE_WORD"),
            GetLine(f"{self.block}1.{field_name}.OFFSET"),
        )

        field_info = BitOutFieldInfo(
            field_type, field_subtype, desc, capture_word, int(offset)
        )

        return field_name, field_info

    def _bit_mux(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        desc, max_delay, labels = yield from _execute_commands(
            self._get_desc(field_name),
            GetLine(f"{self.block}1.{field_name}.MAX_DELAY"),
            GetMultiline(f"*ENUMS.{self.block}.{field_name}"),
        )

        field_info = BitMuxFieldInfo(
            field_type, field_subtype, desc, int(max_delay), labels
        )

        return field_name, field_info

    def _pos_mux(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        desc, labels = yield from _execute_commands(
            self._get_desc(field_name),
            GetMultiline(f"*ENUMS.{self.block}.{field_name}"),
        )
        field_info = PosMuxFieldInfo(field_type, field_subtype, desc, labels)

        return field_name, field_info

    def _table(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        # Ignore the ROW_WORDS attribute as it's new and won't be present on all PandAs,
        # and there's no easy way to try it and catch an error while also running other
        # Get commands at the same time
        table_desc, max_length, fields = yield from _execute_commands(
            self._get_desc(field_name),
            GetLine(f"{self.block}1.{field_name}.MAX_LENGTH"),
            GetMultiline(f"{self.block}1.{field_name}.FIELDS"),
        )

        # Keep track of highest bit index
        max_bit_offset: int = 0

        desc_gets: list[GetLine] = []
        enum_field_gets: list[GetMultiline] = []
        enum_field_names: list[str] = []
        fields_dict: dict[str, TableFieldDetails] = {}
        for field_details in fields:
            # Fields are of the form <bit_high>:<bit_low> <name> <subtype>
            bit_range, name, subtype = field_details.split()
            bit_high_str, bit_low_str = bit_range.split(":")
            bit_high = int(bit_high_str)
            bit_low = int(bit_low_str)

            if bit_high > max_bit_offset:
                max_bit_offset = bit_high

            if subtype == "enum":
                enum_field_gets.append(
                    GetMultiline(f"*ENUMS.{self.block}1.{field_name}[].{name}")
                )
                enum_field_names.append(name)

            fields_dict[name] = TableFieldDetails(subtype, bit_low, bit_high)

            desc_gets.append(GetLine(f"*DESC.{self.block}1.{field_name}[].{name}"))

        # Calculate the number of 32 bit words that comprises one table row
        row_words = max_bit_offset // 32 + 1

        # The first len(enum_field_gets) items are enum labels, type list[str]
        # The second part of the list are descriptions, type str
        labels_and_descriptions = yield from _execute_commands(
            *enum_field_gets, *desc_gets
        )

        for name, labels in zip(
            enum_field_names, labels_and_descriptions[: len(enum_field_gets)]
        ):
            fields_dict[name].labels = labels

        for name, desc in zip(
            fields_dict.keys(), labels_and_descriptions[len(enum_field_gets) :]
        ):
            fields_dict[name].description = desc

        field_info = TableFieldInfo(
            field_type,
            field_subtype,
            table_desc,
            int(max_length),
            fields_dict,
            row_words,
        )

        return field_name, field_info

    def _pos_out(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        desc, capture_labels = yield from _execute_commands(
            self._get_desc(field_name),
            GetMultiline(f"*ENUMS.{self.block}.{field_name}.CAPTURE"),
        )

        field_info = PosOutFieldInfo(field_type, field_subtype, desc, capture_labels)
        return field_name, field_info

    def _ext_out(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        desc, capture_labels = yield from _execute_commands(
            self._get_desc(field_name),
            GetMultiline(f"*ENUMS.{self.block}.{field_name}.CAPTURE"),
        )

        field_info = ExtOutFieldInfo(field_type, field_subtype, desc, capture_labels)
        return field_name, field_info

    def _ext_out_bits(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        desc, bits, capture_labels = yield from _execute_commands(
            self._get_desc(field_name),
            GetMultiline(f"{self.block}.{field_name}.BITS"),
            GetMultiline(f"*ENUMS.{self.block}.{field_name}.CAPTURE"),
        )
        field_info = ExtOutBitsFieldInfo(
            field_type, field_subtype, desc, capture_labels, bits
        )
        return field_name, field_info

    def _no_attributes(
        self, field_name: str, field_type: str, field_subtype: Optional[str]
    ) -> _FieldGeneratorType:
        """Calling this method indicates type-subtype pair is known and
        has no attributes, so only a description needs to be retrieved"""
        desc = yield from self._get_desc(field_name).execute()

        return field_name, FieldInfo(field_type, field_subtype, desc)

    def execute(self) -> ExchangeGenerator[dict[str, FieldInfo]]:
        ex = Exchange(f"{self.block}.*?")
        yield ex
        unsorted: dict[int, tuple[str, FieldInfo]] = {}
        field_generators: list[ExchangeGenerator] = []

        for line in ex.multiline:
            field_name, index, type_subtype = line.split(maxsplit=2)

            field_type: str
            subtype: Optional[str]
            # Append "None" to list below so there are always at least 2 elements
            # so we can always unpack into subtype, even if no split occurs.
            field_type, subtype, *_ = [*type_subtype.split(maxsplit=1), None]

            # Always create default FieldInfo. If necessary we will replace it later
            # with a more type-specific version.
            field_info = FieldInfo(field_type, subtype, None)

            if self.extended_metadata:
                try:
                    # Construct the list of type-specific generators
                    field_generators.append(
                        self._commands_map[(field_type, subtype)](
                            field_name, field_type, subtype
                        )
                    )

                except KeyError:
                    # This exception will be hit if PandA ever defines new types
                    logging.exception(
                        f"Unknown type {(field_type, subtype)} detected for "
                        f"{field_name}, cannot retrieve extended information for it."
                    )
                    # We can assume the new field will have a description though
                    field_generators.append(
                        self._no_attributes(field_name, field_type, subtype)
                    )

            # Keep track of order of fields as returned by PandA. Important for later
            # matching descriptions back to their field.
            unsorted[int(index)] = (field_name, field_info)

        # Dict keeps insertion order, so insert in the order the server said
        fields = {name: field for _, (name, field) in sorted(unsorted.items())}

        if self.extended_metadata is False:
            # Asked to not perform the requests for extra metadata.
            return fields

        field_name_info: tuple[tuple[str, FieldInfo], ...]
        field_name_info = yield from _zip_with_return(field_generators)

        fields.update(field_name_info)

        return fields


class GetPcapBitsLabels(Command):
    """Get the labels for the bit fields in PCAP.

    For example::

        GetPcapBitsLabels() -> {"BITS0" : ["TTLIN1.VAL", "TTLIN2.VAL", ...], ...}
    """

    def execute(self) -> ExchangeGenerator[dict[str, list[str]]]:
        ex = Exchange("PCAP.*?")
        yield ex
        bits_fields = []
        for line in ex.multiline:
            split = line.split()
            if len(split) == 4:
                field_name, _, field_type, field_subtype = split

                if field_type == "ext_out" and field_subtype == "bits":
                    bits_fields.append(f"PCAP.{field_name}")

        exchanges = [Exchange(f"{field}.BITS?") for field in bits_fields]
        yield exchanges
        bits = {field: ex.multiline for field, ex in zip(bits_fields, exchanges)}
        return bits


class ChangeGroup(Enum):
    """Which group of values to ask for ``*CHANGES`` on:
    https://pandablocks-server.readthedocs.io/en/latest/commands.html#system-commands
    """

    #: All the groups below
    ALL = ""
    #: Configuration settings
    CONFIG = ".CONFIG"
    #: Bits on the system bus
    BITS = ".BITS"
    #: Positions
    POSN = ".POSN"
    #: Polled read values
    READ = ".READ"
    #: Attributes (included capture enable flags)
    ATTR = ".ATTR"
    #: Table changes
    TABLE = ".TABLE"
    #: Table changes
    METADATA = ".METADATA"


@dataclass
class GetChanges(Command[Changes]):
    """Get a `Changes` object showing which fields have changed since the last
    time this was called

    Args:
        group: Restrict to a particular `ChangeGroup`
        get_multiline: If `True`, return values of multiline fields in the
            `multiline_values` attribute. Note that this will invoke additional network
            requests.
            If `False` these fields will instead be returned in the `no_value`
            attribute. Default value is `False`.

    For example::

        GetChanges() -> Changes(
            value={"PCAP.TRIG": "PULSE1.OUT"},
            no_value=["SEQ1.TABLE"],
            in_error=["BAD.ENUM"],
            multiline_values={}
        )

        GetChanges(ChangeGroup.ALL, True) -> Changes(
            values={"PCAP.TRIG": "PULSE1.OUT"},
            no_value=[],
            in_error=["BAD.ENUM"],
            multiline_values={"SEQ1.TABLE" : ["1", "2", "3",...]}
        )
    """

    group: ChangeGroup = ChangeGroup.ALL
    get_multiline: bool = False

    def execute(self) -> ExchangeGenerator[Changes]:
        ex = Exchange(f"*CHANGES{self.group.value}?")
        yield ex
        changes = Changes({}, [], [], {})
        multivalue_get_commands: list[tuple[str, GetMultiline]] = []
        for line in ex.multiline:
            if line[-1] == "<":
                if self.get_multiline:
                    field = line[0:-1]
                    multivalue_get_commands.append((field, GetMultiline(field)))
                else:
                    changes.no_value.append(line[:-1])

            elif line.endswith("(error)"):
                changes.in_error.append(line.split(" ", 1)[0])
            else:
                field, value = line.split("=", maxsplit=1)
                changes.values[field] = value

        if self.get_multiline:
            multiline_vals = yield from _execute_commands(
                *[item[1] for item in multivalue_get_commands]
            )

            for field, value in zip(
                [item[0] for item in multivalue_get_commands], multiline_vals
            ):
                assert isinstance(value, list)
                changes.multiline_values[field] = value

        return changes


@dataclass
class GetState(Command[list[str]]):
    """Get the state of all the fields in a PandA that should be saved as a
    list of raw lines that could be sent with `SetState`.

    NOTE: `GetState` may behave unexpectedly if `GetChanges` has previously been
    called using the same client. The caller should use separate clients to avoid
    potential issues.

    For example::

        GetState() -> [
            "SEQ1.TABLE<B"
            "234fds0SDklnmnr"
            ""
            "PCAP.TRIG=PULSE1.OUT",
        ]
    """

    def execute(self) -> ExchangeGenerator[list[str]]:
        # TODO: explain in detail how this works
        # See: references/how-it-works
        attr, config, table, metadata = yield from _execute_commands(
            GetChanges(ChangeGroup.ATTR),
            GetChanges(ChangeGroup.CONFIG),
            GetChanges(ChangeGroup.TABLE),
            GetChanges(ChangeGroup.METADATA),
        )
        # Add the single line values
        line_values = dict(**attr.values, **config.values, **metadata.values)
        state = [f"{k}={v}" for k, v in line_values.items()]
        # Get the multiline values
        multiline_keys, commands = [], []
        for field_name in table.no_value:
            # Get tables as base64
            multiline_keys.append(f"{field_name}<B")
            commands.append(GetMultiline(f"{field_name}.B"))
        for field_name in metadata.no_value:
            # Get metadata as string list
            multiline_keys.append(f"{field_name}<")
            commands.append(GetMultiline(f"{field_name}"))
        multiline_values = yield from _execute_commands(*commands)
        for k, v in zip(multiline_keys, multiline_values):
            state += [k] + v + [""]
        return state


@dataclass
class SetState(Command[None]):
    """Set the state of all the fields in a PandA

    Args:
        state: A list of raw lines as produced by `GetState`

    For example::

        SetState([
            "SEQ1.TABLE<B"
            "234fds0SDklnmnr"
            ""
            "PCAP.TRIG=PULSE1.OUT",
        ])
    """

    state: list[str]

    def execute(self) -> ExchangeGenerator[None]:
        commands: list[Raw] = []
        command_lines: list[str] = []
        for line in self.state:
            command_lines.append(line)
            first_line = len(command_lines) == 1
            if (first_line and not is_multiline_command(line)) or not line:
                # If not a multiline command
                # Or blank line at the end of a multiline command
                commands.append(Raw(command_lines))
                command_lines = []
        returns = yield from _execute_commands(*commands)
        for command, ret in zip(commands, returns):
            if ret != ["OK"]:
                logging.warning(f"command {command.inp} failed with {ret}")



import pytest

from pandablocks.blocking import BlockingClient
from pandablocks.commands import CommandException, Get, Put


def test_blocking_get(dummy_server_in_thread):
    dummy_server_in_thread.send.append("OK =something")
    with BlockingClient("localhost") as client:
        response = client.send(Get("PCAP.ACTIVE"), timeout=1)
    assert response == "something"
    assert dummy_server_in_thread.received == ["PCAP.ACTIVE?"]


def test_blocking_bad_put_raises(dummy_server_in_thread):
    dummy_server_in_thread.send.append("ERR no such field")
    with BlockingClient("localhost") as client:
        with pytest.raises(CommandException) as cm:
            client.send(Put("PCAP.thing", 1), timeout=1)
        assert (
            str(cm.value) == "Put(field='PCAP.thing', value=1) raised error:\n"
            "AssertionError: 'PCAP.thing=1' -> 'ERR no such field'"
        )
    assert dummy_server_in_thread.received == ["PCAP.thing=1"]


def test_blocking_data(
    dummy_server_in_thread,
    slow_dump,
    slow_dump_expected,
):
    dummy_server_in_thread.data = slow_dump
    events = []
    with BlockingClient("localhost") as client:
        for data in client.data(frame_timeout=1):
            events.append(data)
            if len(events) == 8:
                break
    assert slow_dump_expected == events



from collections import OrderedDict
from collections.abc import Iterator

import pytest

from pandablocks.commands import (
    Append,
    ChangeGroup,
    CommandException,
    Get,
    GetBlockInfo,
    GetChanges,
    GetFieldInfo,
    GetLine,
    GetMultiline,
    GetPcapBitsLabels,
    GetState,
    Put,
    SetState,
    is_multiline_command,
)
from pandablocks.connections import (
    ControlConnection,
    DataConnection,
    NoContextAvailable,
)
from pandablocks.responses import (
    BitMuxFieldInfo,
    BitOutFieldInfo,
    BlockInfo,
    Changes,
    Data,
    EnumFieldInfo,
    ExtOutBitsFieldInfo,
    ExtOutFieldInfo,
    FieldInfo,
    PosMuxFieldInfo,
    PosOutFieldInfo,
    ScalarFieldInfo,
    SubtypeTimeFieldInfo,
    TableFieldDetails,
    TableFieldInfo,
    TimeFieldInfo,
    UintFieldInfo,
)
from tests.conftest import STATE_RESPONSES, STATE_SAVEFILE


def get_responses(conn: ControlConnection, received=b""):
    assert not conn.receive_bytes(received)
    return list(conn.responses())


def test_connection_gets_split_value():
    conn = ControlConnection()
    cmd = Get("PCAP.ACTIVE")
    assert conn.send(cmd) == b"PCAP.ACTIVE?\n"
    assert not get_responses(conn, b"OK =1")
    assert get_responses(conn, b"\nAnySpamWeLike") == [(cmd, "1")]


def test_connection_gets_muliline():
    conn = ControlConnection()
    cmd = Get("SEQ1.TABLE")
    assert conn.send(cmd) == b"SEQ1.TABLE?\n"
    assert not get_responses(conn, b"!1048576\n!0\n!10")
    assert get_responses(conn, b"00\n!1000\n.\n") == [
        (cmd, ["1048576", "0", "1000", "1000"])
    ]


def test_get_line():
    conn = ControlConnection()
    cmd = GetLine("PCAP.ACTIVE")
    assert conn.send(cmd) == b"PCAP.ACTIVE?\n"
    assert get_responses(conn, b"OK =1\n") == [(cmd, "1")]


def test_get_line_error_when_multiline():
    conn = ControlConnection()
    cmd = GetLine("PCAP.ACTIVE")
    assert conn.send(cmd) == b"PCAP.ACTIVE?\n"
    assert get_responses(conn, b"!ACTIVE 5 bit_out\n.\n") == [
        (
            cmd,
            ACommandException(
                "GetLine(field='PCAP.ACTIVE') raised error:\n"
                "AssertionError: 'PCAP.ACTIVE?' -> '!ACTIVE 5 bit_out\\n.'"
            ),
        )
    ]


def test_get_line_error_no_ok():
    conn = ControlConnection()
    cmd = GetLine("PCAP.ACTIVE")
    assert conn.send(cmd) == b"PCAP.ACTIVE?\n"
    assert get_responses(conn, b"NOT OK\n") == [
        (
            cmd,
            ACommandException(
                "GetLine(field='PCAP.ACTIVE') raised error:\n"
                "AssertionError: 'PCAP.ACTIVE?' -> 'NOT OK'"
            ),
        )
    ]


def test_get_multiline():
    conn = ControlConnection()
    cmd = GetMultiline("PCAP.*")
    assert conn.send(cmd) == b"PCAP.*?\n"
    assert get_responses(conn, b"!ACTIVE 5 bit_out\n.\n") == [
        (cmd, ["ACTIVE 5 bit_out"])
    ]


def test_get_multiline_error_when_single_line():
    conn = ControlConnection()
    cmd = GetMultiline("PCAP.*")
    assert conn.send(cmd) == b"PCAP.*?\n"

    assert get_responses(conn, b"1\n") == [
        (
            cmd,
            ACommandException(
                "GetMultiline(field='PCAP.*') raised error:\n"
                "AssertionError: 'PCAP.*?' -> '1'"
            ),
        )
    ]


def test_connect_put_single_line():
    conn = ControlConnection()
    cmd = Put("PCAP.TRIG", "PULSE1.OUT")
    assert conn.send(cmd) == b"PCAP.TRIG=PULSE1.OUT\n"
    assert get_responses(conn, b"OK\n") == [(cmd, None)]


class ACommandException(Exception):
    # Compare equal to a CommandException with the same message
    def __eq__(self, other):
        return isinstance(other, CommandException) and other.args == self.args


def test_put_fails_with_single_line_exception():
    conn = ControlConnection()
    cmd = Put("PCAP.blah", "something")
    assert conn.send(cmd) == b"PCAP.blah=something\n"
    assert get_responses(conn, b"ERR No such field\n") == [
        (
            cmd,
            ACommandException(
                "Put(field='PCAP.blah', value='something') raised error:\n"
                "AssertionError: 'PCAP.blah=something' -> 'ERR No such field'"
            ),
        )
    ]


def test_put_fails_with_multiline_exception():
    conn = ControlConnection()
    cmd = Put("PCAP.blah", "something")
    assert conn.send(cmd) == b"PCAP.blah=something\n"
    assert get_responses(conn, b"!This is bad\n!Very bad\n!Don't do this\n.\n") == [
        (
            cmd,
            ACommandException(
                "Put(field='PCAP.blah', value='something') raised error:\n"
                "AssertionError: 'PCAP.blah=something' -> "
                '"!This is bad\\n!Very bad\\n!Don\'t do this\\n."'
            ),
        )
    ]


def test_connect_put_multi_line():
    conn = ControlConnection()
    cmd = Put("SEQ1.TABLE", ["1048576", "0", "1000", "1000"])
    assert conn.send(cmd) == b"SEQ1.TABLE<\n1048576\n0\n1000\n1000\n\n"
    assert get_responses(conn, b"OK\n") == [(cmd, None)]


def test_connect_put_multi_line_bad_list_format():
    """Confirm that an invalid data format raises the expected exception"""
    conn = ControlConnection()
    cmd = Put("SEQ1.TABLE", [1, 2, 3])
    with pytest.raises(TypeError):
        assert conn.send(cmd) == b""


def test_connect_put_no_value():
    """Confirm Put works with no value"""
    conn = ControlConnection()
    cmd = Put("SFP3_SYNC_IN1.SYNC_RESET")
    assert conn.send(cmd) == b"SFP3_SYNC_IN1.SYNC_RESET=\n"
    assert get_responses(conn, b"OK\n") == [(cmd, None)]


def test_connect_append():
    conn = ControlConnection()
    cmd = Append("SEQ1.TABLE", ["1048576", "0", "1000", "1000"])
    assert conn.send(cmd) == b"SEQ1.TABLE<<\n1048576\n0\n1000\n1000\n\n"
    assert get_responses(conn, b"OK\n") == [(cmd, None)]


def test_connect_append_multi_bad_list_format():
    """Confirm that an invalid data format raises the expected exception"""
    conn = ControlConnection()
    cmd = Append("SEQ1.TABLE", [1, 2, 3])
    with pytest.raises(TypeError):
        assert conn.send(cmd) == b""


def test_get_block_info():
    conn = ControlConnection()
    cmd = GetBlockInfo()
    assert conn.send(cmd) == b"*BLOCKS?\n"

    # Respond to first yield, the return from the BLOCKS? command
    assert conn.receive_bytes(b"!PCAP 1\n!LUT 8\n.\n") == b"*DESC.PCAP?\n*DESC.LUT?\n"

    # First of the *DESC.{block}? yields
    assert (
        conn.receive_bytes(b"OK =Description for PCAP field\n") == b""
    )  # No data returned as there's still one outstanding request

    # Create an OrderedDict of the output to test key order - that won't happen
    # with a regular dict
    ordered_dict = OrderedDict(
        [
            ("LUT", BlockInfo(number=8, description="Description for LUT field")),
            ("PCAP", BlockInfo(number=1, description="Description for PCAP field")),
        ]
    )

    # Second and last of the *DESC.{block}? yields - as this is the last response we
    # can call get_responses to also get the overall result
    assert not get_responses(conn)
    assert get_responses(conn, b"OK =Description for LUT field\n") == [
        (
            cmd,
            ordered_dict,
        ),
    ]


def test_get_block_info_skip_description():
    """Test that the skip_description flag causes GetBlockInfo to not retrieve
    descriptions"""
    conn = ControlConnection()
    cmd = GetBlockInfo(skip_description=True)
    assert conn.send(cmd) == b"*BLOCKS?\n"

    ordered_dict = OrderedDict(
        [
            ("PCAP", BlockInfo(number=1, description=None)),
        ]
    )
    # Only a yield for the BLOCKS.* request, not for description as well
    assert get_responses(conn, b"!PCAP 1\n.\n") == [(cmd, ordered_dict)]


def test_get_block_info_error():
    """Test that any errors from *BLOCKS command are correctly reported"""
    conn = ControlConnection()
    cmd = GetBlockInfo()
    assert conn.send(cmd) == b"*BLOCKS?\n"

    # Provide error from PandA server
    assert conn.receive_bytes(b"ERR Cannot read blocks\n") == b""

    assert get_responses(conn) == [
        (
            cmd,
            ACommandException(
                "GetBlockInfo(skip_description=False) raised error:\n"
                "AssertionError: '*BLOCKS?' -> 'ERR Cannot read blocks'"
            ),
        )
    ]


def test_get_block_info_desc_err():
    """Test when the DESC command returns an error"""
    conn = ControlConnection()
    cmd = GetBlockInfo()
    assert conn.send(cmd) == b"*BLOCKS?\n"

    # Respond to first yield, the return from the BLOCKS? command
    assert conn.receive_bytes(b"!PCAP 1\n.\n") == b"*DESC.PCAP?\n"

    # First of the *DESC.{block}? yields
    assert (
        conn.receive_bytes(b"ERR could not get description\n") == b""
    )  # No data returned as there's still one outstanding request

    assert get_responses(conn) == [
        (
            cmd,
            ACommandException(
                "GetBlockInfo(skip_description=False) raised error:\n"
                "AssertionError: '*DESC.PCAP?' -> 'ERR could not get description'"
            ),
        )
    ]


def test_get_fields():
    """Simple test case for GetFieldInfo"""
    conn = ControlConnection()
    cmd = GetFieldInfo("LUT")
    assert conn.send(cmd) == b"LUT.*?\n"

    # First yield, the response to "LUT.*?"
    assert (
        conn.receive_bytes(b"!TYPEA 5 param enum\n!INPA 1 bit_mux\n.\n")
        == b"*DESC.LUT.TYPEA?\n*ENUMS.LUT.TYPEA?\n*DESC.LUT.INPA?\n"
        + b"LUT1.INPA.MAX_DELAY?\n*ENUMS.LUT.INPA?\n"
    )

    # Responses to the 2 *DESC, 2 *ENUM, and MAX_DELAY commands
    responses = [
        b"OK =Source of the value of A for calculation\n",
        b"!Input-Level\n!Pulse-On-Rising-Edge\n.\n",
        b"OK =Input A\n",
        b"OK =10\n",
        b"!TTLIN1.VAL\n!LVDSIN1.VAL\n.\n",
    ]
    for response in responses:
        assert (
            conn.receive_bytes(response) == b""
        )  # Expect no bytes back as none of these trigger further commands

    assert get_responses(conn) == [
        (
            cmd,
            {
                "INPA": BitMuxFieldInfo(
                    type="bit_mux",
                    subtype=None,
                    description="Input A",
                    labels=["TTLIN1.VAL", "LVDSIN1.VAL"],
                    max_delay=10,
                ),
                "TYPEA": EnumFieldInfo(
                    type="param",
                    subtype="enum",
                    description="Source of the value of A for calculation",
                    labels=["Input-Level", "Pulse-On-Rising-Edge"],
                ),
            },
        )
    ]


@pytest.mark.parametrize("gate_duration_name", ["GATE_DURATION", "SAMPLES"])
def test_get_fields_type_ext_out(gate_duration_name):
    """Test for field type == ext_out, ensuring we add .CAPTURE to the end of the
    *ENUMS command"""
    conn = ControlConnection()
    cmd = GetFieldInfo("PCAP")
    assert conn.send(cmd) == b"PCAP.*?\n"

    # First yield, the response to "PCAP.*?"
    request_str = bytes(f"!{gate_duration_name} 9 ext_out samples\n.\n", "utf-8")
    response_str = bytes(
        f"*DESC.PCAP.{gate_duration_name}?\n*ENUMS.PCAP.{gate_duration_name}.CAPTURE?\n",
        "utf-8",
    )
    assert conn.receive_bytes(request_str) == response_str

    # Responses to the *DESC and *ENUM commands
    responses = [
        b"OK =Number of gated samples in the current capture\n",
        b"!No\n!Value\n.\n",
    ]
    for response in responses:
        assert (
            conn.receive_bytes(response) == b""
        )  # Expect no bytes back as none of these trigger further commands

    assert get_responses(conn) == [
        (
            cmd,
            {
                gate_duration_name: ExtOutFieldInfo(
                    type="ext_out",
                    subtype="samples",
                    description="Number of gated samples in the current capture",
                    capture_labels=["No", "Value"],
                )
            },
        )
    ]


def test_get_fields_skip_metadata():
    """Test that the skip_metadata flag causes no description to be retrieved
    for the field"""
    conn = ControlConnection()
    cmd = GetFieldInfo("PCAP", False)
    assert conn.send(cmd) == b"PCAP.*?\n"

    assert conn.receive_bytes(b"!SAMPLES 9 ext_out samples\n.\n") == b""

    assert get_responses(conn) == [
        (
            cmd,
            {"SAMPLES": FieldInfo(type="ext_out", subtype="samples", description=None)},
        )
    ]


def test_get_fields_non_existant_block():
    """Test that querying for an unknown block returns a sensible error"""
    conn = ControlConnection()
    cmd = GetFieldInfo("FOO")
    assert conn.send(cmd) == b"FOO.*?\n"

    # Provide the error string the PandA would provide
    assert conn.receive_bytes(b"ERR No such block\n") == b""

    assert get_responses(conn) == [
        (
            cmd,
            ACommandException(
                "GetFieldInfo(block='FOO', extended_metadata=True) raised error:\n"
                "AssertionError: 'FOO.*?' -> 'ERR No such block'"
            ),
        )
    ]


def test_get_fields_unknown_fields():
    """Test that querying for an unknown field and/or subtype still tries to request a
    description"""
    conn = ControlConnection()
    cmd = GetFieldInfo("PCAP")
    assert conn.send(cmd) == b"PCAP.*?\n"

    # TEST1 has unknown field name, TEST2 has both unknown field and subtype name,
    # TEST3 has known field name but unknown subtype, TEST4 has unknown type but known
    # subtype
    assert (
        conn.receive_bytes(
            b"!TEST1 1 foo\n!TEST2 2 flibble bibble\n!TEST3 3 param unknown\n"
            + b"!TEST4 4 fish uint\n.\n"
        )
        == b"*DESC.PCAP.TEST1?\n*DESC.PCAP.TEST2?\n*DESC.PCAP.TEST3?\n"
        + b"*DESC.PCAP.TEST4?\n"
    )

    responses = [
        b"OK =TEST1 Desc\n",
        b"OK =TEST2 Desc\n",
        b"OK =TEST3 Desc\n",
        b"OK =TEST4 Desc\n",
    ]

    for response in responses:
        assert conn.receive_bytes(response) == b""

    assert get_responses(conn) == [
        (
            cmd,
            {
                "TEST1": FieldInfo(type="foo", subtype=None, description="TEST1 Desc"),
                "TEST2": FieldInfo(
                    type="flibble", subtype="bibble", description="TEST2 Desc"
                ),
                "TEST3": FieldInfo(
                    type="param", subtype="unknown", description="TEST3 Desc"
                ),
                "TEST4": FieldInfo(
                    type="fish", subtype="uint", description="TEST4 Desc"
                ),
            },
        ),
    ]


def idfn(val):
    """helper function to nicely name parameterized test IDs"""
    if isinstance(val, str):
        if val.count("?\n") > 0:
            return ""
        return val
    else:
        return ""


# Table field handled in separate test due to extra round of network calls required
@pytest.mark.parametrize(
    "field_type, field_subtype, requests_responses, expected_field_info",
    [
        (
            "param",
            "uint",
            [
                ("TEST1.TEST_FIELD.MAX?", "OK =10"),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            UintFieldInfo("param", "uint", max_val=10, description="Test Description"),
        ),
        (
            "read",
            "uint",
            [
                ("TEST1.TEST_FIELD.MAX?", "OK =67"),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            UintFieldInfo("read", "uint", max_val=67, description="Test Description"),
        ),
        (
            "write",
            "uint",
            [
                ("TEST1.TEST_FIELD.MAX?", "OK =58"),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            UintFieldInfo("write", "uint", max_val=58, description="Test Description"),
        ),
        (
            "param",
            "int",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("param", "int", description="Test Description"),
        ),
        (
            "read",
            "int",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("read", "int", description="Test Description"),
        ),
        (
            "write",
            "int",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("write", "int", description="Test Description"),
        ),
        (
            "param",
            "scalar",
            [
                ("TEST.TEST_FIELD.UNITS?", "OK =some_units"),
                ("TEST.TEST_FIELD.SCALE?", "OK =0.5"),
                ("TEST.TEST_FIELD.OFFSET?", "OK =8"),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            ScalarFieldInfo(
                "param",
                "scalar",
                units="some_units",
                scale=0.5,
                offset=8,
                description="Test Description",
            ),
        ),
        (
            "read",
            "scalar",
            [
                ("TEST.TEST_FIELD.UNITS?", "OK =some_units"),
                ("TEST.TEST_FIELD.SCALE?", "OK =0.5"),
                ("TEST.TEST_FIELD.OFFSET?", "OK =8"),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            ScalarFieldInfo(
                "read",
                "scalar",
                units="some_units",
                scale=0.5,
                offset=8,
                description="Test Description",
            ),
        ),
        (
            "write",
            "scalar",
            [
                ("TEST.TEST_FIELD.UNITS?", "OK =some_units"),
                ("TEST.TEST_FIELD.SCALE?", "OK =0.5"),
                ("TEST.TEST_FIELD.OFFSET?", "OK =8"),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            ScalarFieldInfo(
                "write",
                "scalar",
                units="some_units",
                scale=0.5,
                offset=8,
                description="Test Description",
            ),
        ),
        (
            "param",
            "bit",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("param", "bit", description="Test Description"),
        ),
        (
            "read",
            "bit",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("read", "bit", description="Test Description"),
        ),
        (
            "read",
            "bit",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("read", "bit", description="Test Description"),
        ),
        (
            "param",
            "action",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("param", "action", description="Test Description"),
        ),
        (
            "read",
            "action",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("read", "action", description="Test Description"),
        ),
        (
            "write",
            "action",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("write", "action", description="Test Description"),
        ),
        (
            "param",
            "lut",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("param", "lut", description="Test Description"),
        ),
        (
            "read",
            "lut",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("read", "lut", description="Test Description"),
        ),
        (
            "write",
            "lut",
            [
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            FieldInfo("write", "lut", description="Test Description"),
        ),
        (
            "param",
            "enum",
            [
                ("*ENUMS.TEST.TEST_FIELD?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            EnumFieldInfo(
                "param", "enum", labels=["VAL1", "VAL2"], description="Test Description"
            ),
        ),
        (
            "read",
            "enum",
            [
                ("*ENUMS.TEST.TEST_FIELD?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            EnumFieldInfo(
                "read", "enum", labels=["VAL1", "VAL2"], description="Test Description"
            ),
        ),
        (
            "write",
            "enum",
            [
                ("*ENUMS.TEST.TEST_FIELD?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            EnumFieldInfo(
                "write", "enum", labels=["VAL1", "VAL2"], description="Test Description"
            ),
        ),
        (
            "param",
            "time",
            [
                ("*ENUMS.TEST.TEST_FIELD.UNITS?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            SubtypeTimeFieldInfo(
                "param",
                "time",
                units_labels=["VAL1", "VAL2"],
                description="Test Description",
            ),
        ),
        (
            "read",
            "time",
            [
                ("*ENUMS.TEST.TEST_FIELD.UNITS?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            SubtypeTimeFieldInfo(
                "read",
                "time",
                units_labels=["VAL1", "VAL2"],
                description="Test Description",
            ),
        ),
        (
            "write",
            "time",
            [
                ("*ENUMS.TEST.TEST_FIELD.UNITS?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            SubtypeTimeFieldInfo(
                "write",
                "time",
                units_labels=["VAL1", "VAL2"],
                description="Test Description",
            ),
        ),
        (
            "time",
            None,
            [
                ("*ENUMS.TEST.TEST_FIELD.UNITS?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            TimeFieldInfo(
                "time",
                None,
                units_labels=["VAL1", "VAL2"],
                description="Test Description",
            ),
        ),
        (
            "bit_out",
            None,
            [
                ("TEST1.TEST_FIELD.CAPTURE_WORD?", "OK =PCAP.BITS1"),
                ("TEST1.TEST_FIELD.OFFSET?", "OK =12"),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            BitOutFieldInfo(
                "bit_out",
                None,
                capture_word="PCAP.BITS1",
                offset=12,
                description="Test Description",
            ),
        ),
        (
            "pos_out",
            None,
            [
                ("*ENUMS.TEST.TEST_FIELD.CAPTURE?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            PosOutFieldInfo(
                "pos_out",
                None,
                capture_labels=["VAL1", "VAL2"],
                description="Test Description",
            ),
        ),
        (
            "ext_out",
            "timestamp",
            [
                ("*ENUMS.TEST.TEST_FIELD.CAPTURE?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            ExtOutFieldInfo(
                "ext_out",
                "timestamp",
                capture_labels=["VAL1", "VAL2"],
                description="Test Description",
            ),
        ),
        (
            "ext_out",
            "samples",
            [
                ("*ENUMS.TEST.TEST_FIELD.CAPTURE?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            ExtOutFieldInfo(
                "ext_out",
                "samples",
                capture_labels=["VAL1", "VAL2"],
                description="Test Description",
            ),
        ),
        (
            "ext_out",
            "bits",
            [
                ("TEST.TEST_FIELD.BITS?", "!BITS1\n!BITS2\n."),
                ("*ENUMS.TEST.TEST_FIELD.CAPTURE?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            ExtOutBitsFieldInfo(
                "ext_out",
                "bits",
                bits=["BITS1", "BITS2"],
                capture_labels=["VAL1", "VAL2"],
                description="Test Description",
            ),
        ),
        (
            "bit_mux",
            None,
            [
                ("TEST1.TEST_FIELD.MAX_DELAY?", "OK =25"),
                ("*ENUMS.TEST.TEST_FIELD?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            BitMuxFieldInfo(
                "bit_mux",
                None,
                max_delay=25,
                labels=["VAL1", "VAL2"],
                description="Test Description",
            ),
        ),
        (
            "pos_mux",
            None,
            [
                ("*ENUMS.TEST.TEST_FIELD?", "!VAL1\n!VAL2\n."),
                ("*DESC.TEST.TEST_FIELD?", "OK =Test Description"),
            ],
            PosMuxFieldInfo(
                "pos_mux",
                None,
                labels=["VAL1", "VAL2"],
                description="Test Description",
            ),
        ),
    ],
    ids=idfn,
)
def test_get_fields_parameterized_type(
    field_type, field_subtype, requests_responses, expected_field_info
):
    """Test every defined field type-subtype pair that has a defined function
    and confirm it creates the expected FieldInfo (or subclass) with the expected
    data"""
    conn = ControlConnection()
    cmd = GetFieldInfo("TEST", extended_metadata=True)
    assert conn.send(cmd) == b"TEST.*?\n"

    if field_subtype is None:
        field_subtype = ""  # Ensure we don't write the literal string "None"

    field_definition_str = f"!TEST_FIELD 1 {field_type} {field_subtype}\n.\n"

    # Split to get individual requests in a list - we don't care about the order
    received_bytes = conn.receive_bytes(field_definition_str.encode())
    received_requests_list = received_bytes.decode().splitlines()

    expected_requests = [x[0] for x in requests_responses]
    responses = [x[1] for x in requests_responses]

    for request in received_requests_list:
        idx = expected_requests.index(request)
        response = responses[idx] + "\n"
        assert conn.receive_bytes(response.encode()) == b""
        expected_requests.pop(idx)
        responses.pop(idx)

    assert (
        not expected_requests
    ), f"Did not receive all expected requests: {expected_requests}"

    assert get_responses(conn) == [
        (
            cmd,
            {"TEST_FIELD": expected_field_info},
        )
    ]


def test_get_fields_type_table():
    """Test for table field type, including descriptions and retrieving enum labels"""
    conn = ControlConnection()
    cmd = GetFieldInfo("SEQ")
    assert conn.send(cmd) == b"SEQ.*?\n"

    assert (
        conn.receive_bytes(b"!TABLE 7 table\n.\n")
        == b"*DESC.SEQ.TABLE?\nSEQ1.TABLE.MAX_LENGTH?\nSEQ1.TABLE.FIELDS?\n"
    )

    assert conn.receive_bytes(b"OK =Sequencer table of lines\n") == b""

    assert conn.receive_bytes(b"OK =16384\n") == b""

    assert (
        conn.receive_bytes(
            b"!15:0 REPEATS uint\n!19:16 TRIGGER enum\n!63:32 POSITION int\n.\n"
        )
        == b"*ENUMS.SEQ1.TABLE[].TRIGGER?\n*DESC.SEQ1.TABLE[].REPEATS?\n"
        b"*DESC.SEQ1.TABLE[].TRIGGER?\n*DESC.SEQ1.TABLE[].POSITION?\n"
    )

    responses = [
        b"!Immediate\n!BITA=0\n.\n",
        b"OK =Number of times the line will repeat\n",
        b"OK =The trigger condition to start the phases\n",
        b"OK =The position that can be used in trigger condition\n",
    ]
    for response in responses:
        assert (
            conn.receive_bytes(response) == b""
        )  # Expect no bytes back as none of these trigger further commands

    assert get_responses(conn) == [
        (
            cmd,
            {
                "TABLE": TableFieldInfo(
                    type="table",
                    subtype=None,
                    description="Sequencer table of lines",
                    max_length=16384,
                    row_words=2,  # Calculated from POSITION field's highest used bit
                    fields={
                        "REPEATS": TableFieldDetails(
                            subtype="uint",
                            bit_low=0,
                            bit_high=15,
                            description="Number of times the line will repeat",
                            labels=None,
                        ),
                        "TRIGGER": TableFieldDetails(
                            subtype="enum",
                            bit_low=16,
                            bit_high=19,
                            description="The trigger condition to start the phases",
                            labels=["Immediate", "BITA=0"],
                        ),
                        "POSITION": TableFieldDetails(
                            subtype="int",
                            bit_low=32,
                            bit_high=63,
                            description=(
                                "The position that can be used in trigger condition"
                            ),
                            labels=None,
                        ),
                    },
                )
            },
        )
    ]


def test_get_pcap_bits_labels():
    """Simple working testcase for GetPcapBitsLabels"""

    # PandA's return data when it receives "PCAP.*?"
    PCAP_RETURN = [
        "!BITS2 12 ext_out bits",
        "!SHIFT_SUM 4 param uint",
        "!BITS0 10 ext_out bits",
        ".",
    ]

    # PandA's return data when it receives "PCAP.BITS2.BITS?"
    BITS2_RETURN = ["!PCOMP2.OUT", "!PGEN1.ACTIVE", "!PGEN2.ACTIVE", "!PULSE1.OUT", "."]

    # PandA's return data when it receives "PCAP.BITS0.BITS?"
    BITS0_RETURN = [
        "!SFP3_SYNC_IN.BIT8",
        "!SFP3_SYNC_IN.BIT9",
        "!SFP3_SYNC_IN.BIT10",
        ".",
    ]

    conn = ControlConnection()
    cmd = GetPcapBitsLabels()
    assert conn.send(cmd) == b"PCAP.*?\n"

    # First yield, requesting response for PCAP.*?
    response_bytes = "\n".join(PCAP_RETURN).encode() + b"\n"
    assert conn.receive_bytes(response_bytes) == b"PCAP.BITS2.BITS?\nPCAP.BITS0.BITS?\n"

    # First of the .BITS? yields
    response_bytes = "\n".join(BITS2_RETURN).encode() + b"\n"
    assert (
        conn.receive_bytes(response_bytes) == b""
    )  # No data returned as there's still one outstanding request

    # Second of the .BITS? yields - as this is the last response we can call
    # get_responses to also get the overall result
    response_bytes = "\n".join(BITS0_RETURN).encode() + b"\n"

    assert not get_responses(conn)
    assert get_responses(conn, response_bytes) == [
        (
            cmd,
            {
                "PCAP.BITS0": [
                    "SFP3_SYNC_IN.BIT8",
                    "SFP3_SYNC_IN.BIT9",
                    "SFP3_SYNC_IN.BIT10",
                ],
                "PCAP.BITS2": [
                    "PCOMP2.OUT",
                    "PGEN1.ACTIVE",
                    "PGEN2.ACTIVE",
                    "PULSE1.OUT",
                ],
            },
        )
    ]


def test_get_pcap_bits_labels_no_bits_fields():
    """Test we get no response when no BITS fields are returned by the PandA"""

    # PandA's return data when it receives "PCAP.*?"
    PCAP_RETURN = [
        "!SHIFT_SUM 4 param uint",
        "!ACTIVE 5 bit_out",
        "!ENABLE 0 bit_mux",
        ".",
    ]
    conn = ControlConnection()
    cmd = GetPcapBitsLabels()
    assert conn.send(cmd) == b"PCAP.*?\n"

    # As there are no BITS fields in the PCAP return, expect no response
    response_bytes = "\n".join(PCAP_RETURN).encode() + b"\n"
    assert conn.receive_bytes(response_bytes) == b""


def test_expected_exception_when_receive_without_send():
    """Test that calling receive_bytes() without first calling send() raises the
    expected exception"""

    conn = ControlConnection()
    with pytest.raises(NoContextAvailable):
        conn.receive_bytes(b"abc\n")


def test_get_changes_values():
    """Test that the `values` field returned from `GetChanges` is correctly populated"""
    conn = ControlConnection()
    cmd = GetChanges()

    assert conn.send(cmd) == b"*CHANGES?\n"

    assert conn.receive_bytes(b"!Field 1=Value1\n!Field2=Other Value\n.\n") == b""

    assert get_responses(conn) == [
        (
            cmd,
            Changes(
                values={"Field 1": "Value1", "Field2": "Other Value"},
                no_value=[],
                in_error=[],
                multiline_values={},
            ),
        )
    ]


def test_get_changes_values_empty():
    """Test that the `values` field returned from `GetChanges` is correctly populated
    when the value of a field is empty"""
    conn = ControlConnection()
    cmd = GetChanges()

    assert conn.send(cmd) == b"*CHANGES?\n"

    assert conn.receive_bytes(b"!Field 1=\n.\n") == b""

    assert get_responses(conn) == [
        (
            cmd,
            Changes(
                values={"Field 1": ""},
                no_value=[],
                in_error=[],
                multiline_values={},
            ),
        )
    ]


def test_get_changes_no_value():
    """Test that the `no_value` field returned from `GetChanges` is correctly
    populated"""
    conn = ControlConnection()
    cmd = GetChanges()

    assert conn.send(cmd) == b"*CHANGES?\n"

    assert conn.receive_bytes(b"!Field 1<\n!Field2<\n.\n") == b""

    assert get_responses(conn) == [
        (
            cmd,
            Changes(
                values={},
                no_value=["Field 1", "Field2"],
                in_error=[],
                multiline_values={},
            ),
        )
    ]


def test_get_changes_error():
    """Test that the `error` field returned from `GetChanges` is correctly populated"""
    conn = ControlConnection()
    cmd = GetChanges()

    assert conn.send(cmd) == b"*CHANGES?\n"

    assert conn.receive_bytes(b"!Field1 (error)\n!Field2 (error)\n.\n") == b""

    assert get_responses(conn) == [
        (
            cmd,
            Changes(
                values={},
                no_value=[],
                in_error=["Field1", "Field2"],
                multiline_values={},
            ),
        )
    ]


def test_get_changes_multiline():
    """Test that the `multiline_values` field returned from `GetChanges` is correctly
    populated"""
    conn = ControlConnection()
    cmd = GetChanges(ChangeGroup.ALL, True)

    assert conn.send(cmd) == b"*CHANGES?\n"

    assert conn.receive_bytes(b"!Field1<\n.\n") == b"Field1?\n"

    assert get_responses(conn, b"!Val1\n!Val2\n.\n") == [
        (
            cmd,
            Changes(
                values={},
                no_value=[],
                in_error=[],
                multiline_values={"Field1": ["Val1", "Val2"]},
            ),
        )
    ]


def test_get_changes_multiline_no_values():
    """Test that the `multiline_values` field returned from `GetChanges` is correctly
    populated when the table in question has no values"""
    conn = ControlConnection()
    cmd = GetChanges(ChangeGroup.ALL, True)

    assert conn.send(cmd) == b"*CHANGES?\n"

    assert conn.receive_bytes(b"!Field1<\n.\n") == b"Field1?\n"

    assert get_responses(conn, b".\n") == [
        (
            cmd,
            Changes(
                values={},
                no_value=[],
                in_error=[],
                multiline_values={"Field1": []},
            ),
        )
    ]


def test_get_changes_multiline_no_multiline_fields():
    """Test retrieving multiline fields when none are defined returns expected empty
    multiline_values field."""
    conn = ControlConnection()
    cmd = GetChanges(ChangeGroup.ALL, True)

    assert conn.send(cmd) == b"*CHANGES?\n"

    assert conn.receive_bytes(b"!Field1=Value1\n.\n") == b""

    assert get_responses(conn) == [
        (
            cmd,
            Changes(
                values={"Field1": "Value1"},
                no_value=[],
                in_error=[],
                multiline_values={},
            ),
        )
    ]


def test_save():
    conn = ControlConnection()
    cmd = GetState()
    assert (
        conn.send(cmd)
        == b"*CHANGES.ATTR?\n*CHANGES.CONFIG?\n*CHANGES.TABLE?\n*CHANGES.METADATA?\n"
    )
    response_bytes = "\n".join(STATE_RESPONSES).encode() + b"\n"
    assert (
        conn.receive_bytes(response_bytes[:107])
        == b"Table.B?\nMultiLineMeta1?\nMultiLineMeta2?\n"
    )
    assert not get_responses(conn)
    assert get_responses(conn, response_bytes[107:]) == [(cmd, STATE_SAVEFILE)]


def test_load():
    conn = ControlConnection()
    cmd = SetState(STATE_SAVEFILE)
    assert conn.send(cmd) == ("\n".join(STATE_SAVEFILE)).encode() + b"\n"
    response_bytes = "\n".join(["OK"] * 7).encode() + b"\n"
    assert get_responses(conn, response_bytes) == [(cmd, None)]


def get_data(it: Iterator[bytes]) -> Iterator[Data]:
    conn = DataConnection()
    assert conn.connect(scaled=True) == b"XML FRAMED SCALED\n"
    for received in it:
        yield from conn.receive_bytes(received)


def test_slow_data_collection(slow_dump, slow_dump_expected):
    responses = list(get_data(slow_dump))
    assert slow_dump_expected == responses


def test_fast_data_collection(fast_dump, fast_dump_expected):
    responses = list(get_data(fast_dump))
    assert fast_dump_expected == responses
    assert responses[2].column_names == (
        "PCAP.BITS2.Value",
        "COUNTER1.OUT.Min",
        "COUNTER1.OUT.Max",
        "COUNTER3.OUT.Value",
        "PCAP.TS_START.Value",
        "COUNTER1.OUT.Mean",
        "COUNTER2.OUT.Mean",
    )
    assert responses[2].data["COUNTER1.OUT.Mean"] == pytest.approx(range(1, 11))


def test_is_multiline_command():
    assert is_multiline_command("SEQ.TABLE<")
    assert is_multiline_command("SEQ.TABLE<B")
    assert is_multiline_command("*METADATA.DESIGN<")
    assert not is_multiline_command("SEQ.TABLE?")
    assert not is_multiline_command("*METADATA.DESIGN?")
    assert not is_multiline_command("*METADATA.DESIGN=B<B<B<B?")



from collections import deque
from pathlib import Path
from unittest.mock import patch

import h5py
import pytest
from click.testing import CliRunner

from pandablocks import cli
from pandablocks.hdf import GATE_DURATION_FIELD, SAMPLES_FIELD, HDFDataOverrunException
from tests.conftest import (
    STATE_RESPONSES,
    STATE_SAVEFILE,
    DummyServer,
    assert_all_data_in_hdf_file,
)


@pytest.mark.parametrize("samples_name", [GATE_DURATION_FIELD, SAMPLES_FIELD])
def test_writing_fast_hdf(
    samples_name,
    dummy_server_in_thread: DummyServer,
    raw_dump,
    raw_dump_no_duration,
    tmp_path,
):
    dummy_server_in_thread.send.append("OK")
    if samples_name == GATE_DURATION_FIELD:
        dummy_server_in_thread.data = raw_dump
    else:
        dummy_server_in_thread.data = raw_dump_no_duration

    runner = CliRunner()
    result = runner.invoke(
        cli.cli, ["hdf", "localhost", str(tmp_path / "%d.h5"), "--arm"]
    )
    assert result.exit_code == 0
    hdf_file = h5py.File(tmp_path / "1.h5", "r")
    assert list(hdf_file) == [
        "COUNTER1.OUT.Max",
        "COUNTER1.OUT.Mean",
        "COUNTER1.OUT.Min",
        "COUNTER2.OUT.Mean",
        "COUNTER3.OUT.Value",
        "PCAP.BITS2.Value",
        samples_name,
        "PCAP.TS_START.Value",
    ]
    assert dummy_server_in_thread.received == ["*PCAP.ARM="]
    assert_all_data_in_hdf_file(hdf_file, samples_name)


def test_writing_overrun_hdf(
    dummy_server_in_thread: DummyServer, overrun_dump, tmp_path
):
    dummy_server_in_thread.send.append("OK")
    dummy_server_in_thread.data = [overrun_dump]
    runner = CliRunner()
    result = runner.invoke(
        cli.cli, ["hdf", "localhost", str(tmp_path / "%d.h5"), "--arm"]
    )
    assert result.exit_code == 1
    assert isinstance(result.exception, HDFDataOverrunException)
    hdf_file = h5py.File(tmp_path / "1.h5", "r")
    assert_all_data_in_hdf_file(hdf_file, "PCAP.GATE_DURATION.Value")


class MockInput:
    def __init__(self, *commands: str):
        self._commands = deque(commands)

    def __call__(self, prompt):
        assert prompt == cli.PROMPT
        try:
            return self._commands.popleft()
        except IndexError as err:
            raise EOFError() from err


def test_interactive_simple(dummy_server_in_thread, capsys):
    mock_input = MockInput("PCAP.ACTIVE?", "SEQ1.TABLE?")
    dummy_server_in_thread.send += ["OK =0", "!1\n!2\n!3\n!4\n."]
    with patch("pandablocks._control.input", side_effect=mock_input):
        runner = CliRunner()
        result = runner.invoke(cli.cli, ["control", "localhost", "--no-readline"])
        assert result.exit_code == 0
        assert result.output == "OK =0\n!1\n!2\n!3\n!4\n.\n\n"


def test_save(dummy_server_in_thread: DummyServer, tmp_path: Path):
    dummy_server_in_thread.send += STATE_RESPONSES
    runner = CliRunner()
    path = tmp_path / "saved_state"
    result = runner.invoke(cli.cli, ["save", "localhost", str(path)])
    assert result.exit_code == 0

    with path.open("r") as stream:
        results = stream.read().splitlines()

    assert results == STATE_SAVEFILE


def test_load(dummy_server_in_thread: DummyServer, tmp_path: Path):
    dummy_server_in_thread.send += ["OK"] * 10
    runner = CliRunner()
    path = tmp_path / "saved_state"

    with path.open("w") as stream:
        for line in STATE_SAVEFILE:
            stream.write(line + "\n")

    result = runner.invoke(cli.cli, ["load", "localhost", str(path)])
    assert result.exit_code == 0, result.exc_info

    assert dummy_server_in_thread.received == STATE_SAVEFILE


def test_load_tutorial(dummy_server_in_thread: DummyServer, tmp_path: Path):
    dummy_server_in_thread.send += ["OK"] * 10000
    runner = CliRunner()

    with cli.TUTORIAL.open("r") as stream:
        commands = stream.read().splitlines()

    result = runner.invoke(cli.cli, ["load", "localhost", "--tutorial"])
    assert result.exit_code == 0, result.exc_info

    assert dummy_server_in_thread.received == commands
